# Copyright 2011-2015 Canonical Ltd
#
from ConfigParser import ConfigParser
from argparse import ArgumentParser
from launchpadlib.launchpad import Launchpad
from lp2kanban.cards2workitems import (
    convert_cards_to_work_items,
    get_blueprint_linked_cards,
    update_blueprints_from_work_items,
    )
from lp2kanban.kanban import (
    BRANCH_MP_REGEX,
    RESOURCE_TYPE_LINK_SERIES,
    LeankitKanban,
    LeankitTaskBoard,
    Record,
    )
import os.path
import re
import sys
from traceback import print_exc
import logging

TITLE_MARKER = "sync:"
DESCRIPTION_MARKER = "Synced with Launchpad"
NO_SYNC_MARKER = "(no-sync)"


def get_arg_parser():
    parser = ArgumentParser(
        description="Update a Kanban board from Launchpad data.")
    parser.add_argument(
        '-c', '--config', default='config.ini',
        help="LeankitKanban configuration file.")
    parser.add_argument(
        '-e', '--email', required=False, default=None,
        help="The Kanban user's registered email address")
    parser.add_argument(
        '-p', '--password', required=False, default=None,
        help="The user's Kanban password")
    parser.add_argument(
        '-b', '--board', required=False,
        help="The Kanban board (a title).")
    parser.add_argument(
        '-l', '--list', action='store_const', const=True, default=False,
        help="List available kanban boards.")
    parser.add_argument(
        '-d', '--debug', action='store_const', const=True, default=False,
        help="Turn on debugging for HTTP traffic")
    parser.add_argument(
        '-B', '--bug', required=False, default=None,
        help="Only operate on this bug.")
    parser.add_argument(
        '-C', '--card', required=False, default=None,
        help="Only operate on this card, can obtain by looking at URL.")
    return parser


def parse_config(config):
    """Parse the kanban configuration from `config`.

    GLOBAL section is used for global settings for accessing the kanban
    board.  DEFAULT contains the default values which are included in
    each of the board sections automatically.

    Other sections should be named after the board title that they
    are configuring.
    """
    globals = {}
    for key, value in config.items('GLOBAL'):
        globals[key] = value

    boards = {}
    for section in config.sections():
        if section in ['GLOBAL']:
            continue
        boards[section] = {}
        for key in config.defaults():
            boards[section][key] = config.get(section, key)
        for key in config.options(section):
            boards[section][key] = config.get(section, key)
    return globals, boards


def parse_groups_config(config):
    """Parse the groups configuration file from `config`.

    The groups config is an ini file, where the sections are the group
    names, and the `members` attribute of each group is a comma
    separated list of Launchpad names.

    Returns a dict mapping a Launchpad name to a group name.
    """
    lp_to_group = {}
    for group in config.sections():
        if not config.has_option(group, "members"):
            continue
        for member in config.get(group, "members").split(","):
            lp_to_group[member.strip()] = group
    return lp_to_group


class LaunchpadUsersForBoard(object):
    """Find Launchpad user accounts for board users and keep them cached.

    Provides three dicts:
     * lp_to_kanban which uses Launchpad account names as keys and points to
       LeankitUser records.
     * kanban_to_lp which uses LeankitUser.id as keys pointing to actual
       Launchpad user records.
     * lp_to_group which uses Launchpad account names as keys and points to
       group names.
    """
    def __init__(self, kanban_to_lp=None, lp_to_kanban=None, lp_to_group=None):
        if kanban_to_lp is None:
            kanban_to_lp = {}
        self.kanban_to_lp = kanban_to_lp
        if lp_to_kanban is None:
            lp_to_kanban = {}
        self.lp_to_kanban = lp_to_kanban
        if lp_to_group is None:
            lp_to_group = {}
        self.lp_to_group = lp_to_group

    def set_up_users(self, lp, board, config):
        """Set up the dicts this object provides."""
        for user in board.users:
            lp_user = lp.people.getByEmail(email=user)
            if lp_user is None:
                print (
                    "WARNING: There is no matching Launchpad user for kanban"
                    " user %s." % (user,))
                continue
            self.lp_to_kanban[lp_user.name] = board.users[user]
            self.kanban_to_lp[board.users[user].id] = lp_user
        groups_config_file = config.get("groups_config_file")
        if groups_config_file:
            assert os.path.exists(groups_config_file), (
                "Groups config file not found: " + groups_config_file)
            groups_config = ConfigParser()
            groups_config.read(groups_config_file)
            self.lp_to_group.update(parse_groups_config(groups_config))

        self.feature_lanes = {}
        new_lane =  config.get("new_lanes")
        if not new_lane:
            return
        if "${group}" in new_lane:
            # Build squad feature lanes to allow for squad-$group tags
            for group in set(self.lp_to_group.values()):
                lane_path = new_lane.replace("${group}", group)
                lane = board.getLaneByPath(lane_path)
                group_tag = "squad-%s" % group.lower().replace(" ","-")
                self.feature_lanes[group_tag] = lane
        for card in board.getFeatureCards():
            if not card.tags:
                print ("Info: feature card '%s' has no tags. New bugs tagged "
                       "for this feature will not be moved to the proper "
                       "new_lane." % card.title)
                continue
            if "${group}" in new_lane:
                lp_user = self.kanban_to_lp.get(card.assigned_user_id)
                if not lp_user:
                    print ("Info: feature card '%s' has no assigned user. New "
                           "bugs tagged for this feature will not be moved to "
                           "the proper new_lane." % card.title)
                    continue
                lane_path = new_lane.replace(
                    "${group}", self.lp_to_group[lp_user.name])
            else:
                lane_path = new_lane
            for tag in card.tags.split(","):
                if tag != "no-sync":
                    self.feature_lanes[tag.strip()] = board.getLaneByPath(
                        lane_path)


ORDERED_STATUSES = {'Rejected': 0,
                    'Work in progress': 0,
                    'Needs review': 1,
                    'Approved': 2,
                    'Merged': 2,
                    }
REVIEW_STATUS = 'In Review'
TRANSFORM_STATUS = {'Rejected': 'In Progress',
                    'Work in progress': 'In Progress',
                    'Needs review': REVIEW_STATUS,
                    'Approved': 'Approved',
                    'Merged': 'Merged',
                    }
IGNORE_STATUSES = ('Superseded',)
FINAL_STATUSES = ('Approved', 'Merged')

def get_lp_bug(card, launchpad):
    """Return the launchpad bug resource for a card if that bug exists.

    Return None if external_card_id is not an int or the bug isn't accessible.
    """
    if not card.external_card_id:
        return None
    try:
        bug_id = int(card.external_card_id)
    except ValueError:
        print "Invalid external id (%s) for card '%s'" % (
            card.external_card_id, card.title)
        return None
    try:
        lp_bug = launchpad.bugs[bug_id]
    except KeyError:
        print (
            "   * %d is not accessible "
            "(does not exist or forbidden)." % (bug_id))
        return None
    return lp_bug


def move_cards_to_feature_taskboard(feature_card, feature_bug, source_lanes):
    """Move all cards related to subtasks of the feature_card.

    This will only move cards which are in the landing_lanes or deploy_lanes.
    @param: feature_card: The card to which all related tasks will be moved.
    @param: feature_bug: The bug for the feature. Any card that points to a
        branch linked to this bug in launchpad will be considered a move
        candidate.
    @param: source_lanes: A list of lane paths. Only cards in these lanes will
        be moved into the taskboard.
    """
    linked_cards = get_cards_for_feature(feature_card, feature_bug=feature_bug)
    for card in linked_cards:
        if card.type.name == "Feature":  # We don't want to move feature cards
             continue
        card_path = card.lane.path
        if card_path in source_lanes:
            print "Moving to taskboard '{}' card-{}".format(
                feature_card.title, card.title)
            card.moveToTaskBoard(feature_card)
        else:
            print (
                "Card in progress, not moving to taskboard '%s' card-%s." %
                   (feature_card.title, card.title))


def get_cards_for_feature(feature_card, feature_bug=None, card_type=None):
    """Return all cards related to the provided feature_card.

    Related cards meet one of two criteria:
       1. The card in kanban has a tag that matches the tag on the feature
          card.
       2. The card's external_system_url is a branch which is linked to a
          the feature bug for this feature.

    @param feature_card: The feature LeankitCard to associate
    @param feature_bug: The Launchpad bug resource linked to this feature_card
    @param card_type: only return cards this type: Branch, Feature, Task, ...
    """
    feature_tags = set()
    linked_cards = set()
    if feature_card.tags:
        feature_tags = set(feature_card.tags.split(","))

    branch_links = []
    if feature_bug:
        branch_links = [
            linked_branch.branch.web_link
            for linked_branch in feature_bug.linked_branches]

    for card in feature_card.lane.board.cards:
        if card_type and card.type.name != card_type:
             continue
        if card.tags:
            card_tags = set(card.tags.split(","))
            if feature_tags.intersection(card_tags):
                linked_cards.add(card)
        if branch_links and card.external_system_url:
            match = BRANCH_MP_REGEX.match(card.external_system_url)
            if not match:
                continue
            # Drop any merge content from branch
            branch_name = match.group("branch")
            url = "https://code.launchpad.net/{}".format(branch_name)
            if url in branch_links:
                linked_cards.add(card)
    return linked_cards


def get_branch_info(branches):
    """Returns a bug's status based on the given branches.

    Result will be one of "Approved", "Merged", "In Review", "In
    Progress", or None.  The meanings of the result values should be
    obvious with the addition that None signifies that no information
    could be gleaned from the branches.
    """
    info = Record(status=None, target=None, mp_url=None, mp_type=None)

    def _get_mp_info():
        mps = []
        has_inprogress_branches = False
        for branch in branches:
            if hasattr(branch, 'branch') :
                branch = branch.branch
            for mp in branch.landing_targets:
                mp_info = Record(rank=None, status=None, mp=None)
                status = mp.queue_status
                if status in IGNORE_STATUSES:
                    continue
                if status not in FINAL_STATUSES:
                    has_inprogress_branches = True
                mp_info.rank = ORDERED_STATUSES.get(status, 1)
                mp_info.status = TRANSFORM_STATUS.get(status, REVIEW_STATUS)
                mp_info.mp = mp
                mps.append(mp_info)
        if len(mps) == 0:
            return None
        mps.sort(key=lambda x:(x.rank, x.mp.date_created), reverse=True)
        if not has_inprogress_branches:
            return mps[0]  # Return highest-ranked and most-recent MP
        for mp in mps:
            if mp.status in FINAL_STATUSES:
                continue
            return mp  # Return highest-ranked non-final MP

    def _get_mp_link(mp):
        regex = re.compile('(https://codereview.appspot.com/\d+)')
        match = regex.search(mp.description or '')
        if match:
            # An appspot Rietveld URL was found.  Use it.
            url = match.group(0)
            mp_type = 'Rietveld'
        else:
            url = mp.web_link
            mp_type = 'LP Merge Proposal'
        return (url, mp_type)

    mp_info = _get_mp_info()
    if mp_info is None:
        return info
    mp = mp_info.mp
    info.status = mp_info.status
    info.target = mp.target_branch.name
    info.mp_url, info.mp_type = _get_mp_link(mp)
    return info


def should_sync_card(card, conf={}):
    """Check to see if the card should be synced.

    `conf` is the kanban board configuration as a dict, and should contain
    an option `autosync` set to 'on' or 'active-projects' if all cards
    with external card IDs should be synced unless their description or title
    contain NO_SYNC_MARKER.
    With autosync off, if the card title starts with TITLE_MARKER or the card
    description contains DESCRIPTION_MARKER then return True.

    If `conf` has the option `sync_cards` set the 'off', False is returned.
    """
    if conf.get('sync_cards', 'off') == 'off':
        return False
    logging.debug("Should I sync? - {}".format(card.title))
    no_sync_lanes = [
        lane_path.strip()
        for lane_path in conf.get("no_sync_lanes", "").split(",")]
    for no_sync_lane_path in no_sync_lanes:
        if no_sync_lane_path and card.lane.path.endswith(no_sync_lane_path):
            logging.debug("Not syncing, {} in no_sync_lanes".format(
                card.lane.path))
            return False
    if conf.get('autosync', None) in ['on', 'active-projects']:
        has_id = (card.external_card_id is not None and
                  card.external_card_id.strip() != '')
        has_branch = (card.external_system_url is not None and
                      BRANCH_MP_REGEX.match(card.external_system_url))
        if NO_SYNC_MARKER in card.title or NO_SYNC_MARKER in card.description:
            return False
        return (has_id or has_branch)
    else:
        return (card.title.startswith(TITLE_MARKER) or
                DESCRIPTION_MARKER in card.description)


def should_move_card(card, conf={}):
    """Check to see if a card is eligible to be moved.

    The configuration parameter `move_cards` must be enabled and the card must
    be in an active portion of the board.
    """
    if card.type.name == "Feature":    # Don't move feature cards at all
        return False
    if card.tags and "no-sync" in set(card.tags.split(",")):
        return False
   
    enabled = conf.get('move_cards', 'off') == 'on'
    active = not card.lane.board.is_archived
    return enabled and active


NEW_BUG_STATUSES = [
    "New", "Confirmed", "Triaged", "Incomplete"
]

NOOP_BUG_STATUSES = [
]
IN_PROGRESS_BUG_STATUSES = [
    "In Progress", "Fix Committed",
    ]
DONE_FIX_BUG_STATUSES = [
    "Fix Released",
]
DONE_NOFIX_BUG_STATUSES = [
    "Opinion", "Invalid", "Won't Fix", "Expired"]

ORDERED_BUG_STATUSES = (
    NEW_BUG_STATUSES + NOOP_BUG_STATUSES + IN_PROGRESS_BUG_STATUSES +
    DONE_FIX_BUG_STATUSES + DONE_NOFIX_BUG_STATUSES)


class CardStatus:
    """Card is consider to be in only one of these statuses."""
    NEW = 0
    CODING = 1
    REVIEW = 2
    LANDING = 3
    QA = 4
    DEPLOY = 5
    DOWNTIME_DEPLOY = 6
    DONE_FIX = 7
    DONE_NOFIX = 8


# Configuration variables for individual card statuses.
ConfigOptionByStatus = {
    CardStatus.NEW: 'new_lanes',
    CardStatus.CODING: 'coding_lanes',
    CardStatus.REVIEW: 'review_lanes',
    CardStatus.LANDING: 'landing_lanes',
    CardStatus.QA: 'qa_lanes',
    CardStatus.DEPLOY: 'deploy_lanes',
    CardStatus.DOWNTIME_DEPLOY: 'downtime_deploy_lanes',
    CardStatus.DONE_FIX: 'done_fix_lanes',
    CardStatus.DONE_NOFIX: 'done_nofix_lanes',
    }


def get_new_bugs(board, bugs):
    """Return the bugs that aren't already in the board."""
    existing_bugs = set()
    for card in board.cards:
        if card.external_card_id:
            try:
                bugid = int(card.external_card_id)
            except ValueError:
                print "Invalid external id (%s) for card '%s'" % (
                    card.external_card_id, card.title)
                continue
            existing_bugs.add(bugid)
    for bug in bugs:
        if bug.id not in existing_bugs:
            yield bug


def get_card_status(bug_status, bug_tags, branch_info):
    """Return the status of the card as one of CardStatus values."""
    # status of None means card needs not to be moved.
    status = None
    branch_card_status = None
    if branch_info.status:
        if branch_info.status == 'In Progress':
            branch_card_status = CardStatus.CODING
        elif branch_info.status == 'In Review':
            branch_card_status = CardStatus.REVIEW
        elif branch_info.status == 'Approved':
            branch_card_status = CardStatus.REVIEW
        elif branch_info.status == 'Merged':
            branch_card_status = CardStatus.LANDING
        if bug_status is None:
            return branch_card_status
    if bug_status in NEW_BUG_STATUSES:
        status = CardStatus.NEW
    elif bug_status in IN_PROGRESS_BUG_STATUSES:
        status = CardStatus.CODING
        if branch_card_status in (None, CardStatus.LANDING):
            if bug_status == 'Fix Committed':
                if 'qa-needstesting' in bug_tags:
                    status = CardStatus.QA
                else:
                    if branch_info.target == 'db-devel':
                        status = CardStatus.DOWNTIME_DEPLOY
                    else:
                        status = CardStatus.DEPLOY
            elif branch_card_status is not None:
                status = CardStatus.LANDING
        elif branch_card_status:
            return branch_card_status
            
    elif bug_status in DONE_FIX_BUG_STATUSES:
        status = CardStatus.DONE_FIX
    elif bug_status in DONE_NOFIX_BUG_STATUSES:
        status = CardStatus.DONE_NOFIX
    return status


def is_bug_in_projects(lp_bug, all_projects):
    """Return True if bug is linked to one of the configured projects"""
    for task in lp_bug.bug_tasks_collection:
        if task.target in all_projects:
            return True
        if (hasattr(task.target, "project") and
            task.target.project in all_projects):
            return True
    return False


def get_bug_status(lp_bug, all_projects, all_users):
    lowest_status = None
    matches = False
    assignees = []
    tasks = lp_bug.bug_tasks_collection
    for task in tasks:
        # Ignore bug task status for series targets
        if (task.target.resource_type_link == RESOURCE_TYPE_LINK_SERIES and
            task.target.project in all_projects):
            continue  # Skip series task status for our projects
        if (task.assignee is not None and
            task.assignee.name not in all_users):
            continue  # Skip non-users tasks
        if (task.target in all_projects or
            task.assignee is not None):
            matches = True
            if (lowest_status is None or
                (ORDERED_BUG_STATUSES.index(task.status) <
                 ORDERED_BUG_STATUSES.index(lowest_status))):
                lowest_status = task.status
            if task.assignee is not None:
                assignees.append(task.assignee)
    branch_info = get_branch_info(lp_bug.linked_branches)
    status = get_card_status(lowest_status, lp_bug.tags, branch_info)
    return matches, status, assignees, branch_info.mp_url, branch_info.mp_type


def find_card_target_lane_omnidirectional(lane, path):
    """Find the target lane for the path using the "omnidirectional" algorithm.

    The "omnidirectional" algorithm searches for the target lane from
    the beginning of the board, no matter what the current lane is.
    """
    for possible_target_lane in lane.board.lanes.values():
        if possible_target_lane.path == path:
            return possible_target_lane



def find_card_target_lane_next(lane, path):
    """Find the target lane for the path using the "next" algorithm.

    The "next" algorithm searches only the current lane's next lanes.
    """
    if lane.path.endswith(path):
        return lane
    else:
        next_lanes = lane.getNextLanes()
        if next_lanes is None:
            return None

        for next_lane in next_lanes:
            target_lane = find_card_target_lane_next(next_lane, path)
            if target_lane is not None:
                return target_lane
        return None


def create_cards_in_project(
        board, launchpad_project, bug_tag, cardtype_name=None,
        cardlane_path=None, feature_lanes=None):
    """Create cards for bugs tagged in Launchpad.

    If a bug is tagged with the given tag, a new card will be created in
    the board and the tag will be removed from the bug.
    """
    print "   Creating new cards for %s:" % launchpad_project.name
    if not feature_lanes:
        feature_lanes = {}
    new_cardtype = board.default_cardtype.id
    if cardtype_name is not None:
        for type_id, cardtype in board.cardtypes.items():
            if cardtype.name == cardtype_name:
                new_cardtype = type_id
    kanban_bugtasks = launchpad_project.searchTasks(tags=[bug_tag])
    kanban_bugs = get_new_bugs(
        board, [bugtask.bug for bugtask in kanban_bugtasks])
    if cardlane_path is not None:
        next_lane = board.getLaneByPath(cardlane_path)
    else:
        next_lane = board.root_lane
        while next_lane.child_lanes:
            next_lane = next_lane.child_lanes[0]
    for bug in kanban_bugs:
        feature_lane = None
        for tag in bug.tags:
            feature_lane = feature_lanes.get(tag, None)
            if feature_lane:  # Found a matching feature for this tag
                card = feature_lane.addCard()
                break
        if not feature_lane:
            card = next_lane.addCard()

        bug_msg = "    Creating card for bug %s in %s: %%s" % (
            bug.id, card.lane.title)
        card.title = bug.title
        # Leankit empirically has a max description length of 19997 bytes
        card.description = bug.description[:19500]
        card.external_card_id = str(bug.id)
        card.type_id = new_cardtype
        saved = card.save()
        # Only update bug tags if card was successfully saved.
        if saved is not None:
            print bug_msg % "DONE"
            bug.tags = [tag for tag in bug.tags if tag != bug_tag]
            bug.lp_save()
        else:
            print bug_msg % "ERROR"


def get_lp(config):
    """Get a lp instance for a logged in user or anonymous access.

    If 'credentials_file' is specified in the board configuration then an
    instance of Launchpad for the user is returned, otherwise anonymous access
    is used.
    """
    credentials_file = config.get('credentials_file')
    launchpadlib_dir = config.get('launchpadlib_dir')
    if credentials_file and os.path.exists(credentials_file):
        lp = Launchpad.login_with(
            'lp2kanban', config['lp_server'], version="devel",
            credentials_file=credentials_file,
            launchpadlib_dir=launchpadlib_dir)
        print "   Logged in as", lp.me.name
    else:
        lp = Launchpad.login_anonymously(
            'lp2kanban', config['lp_server'], version="devel",
            launchpadlib_dir=launchpadlib_dir)
        print ("   Accessing Launchpad anonymously.  "
               "Private bugs will not be synced.")
    return lp


def get_lp_branch_name(url):
    """Return the branch name from either an MP or branch url provided"""
    match = BRANCH_MP_REGEX.match(url)
    if match:
        return match.group("branch")
    return None


def get_lp_data(board, bconf):
    """Given a board and config, return lp, lp projects and lp users."""
    lp = get_lp(bconf)
    lp_users = LaunchpadUsersForBoard()
    all_projects = []
    project_group = bconf.get('project_group')
    if project_group:
        launchpad_project_group = lp.project_groups[project_group]
        # These are sometimes very slow (and basically take a considerable
        # chunk of the runtime).  We should optimize this so it's done only
        # once per run instead of once per board being synced.
        all_projects = list(launchpad_project_group.projects_collection)
    project_names = bconf.get('projects')
    if project_names:
        for name in project_names.split(','):
            project = lp.projects[name.strip()]
            if project:
                all_projects.append(project)
    lp_users.set_up_users(lp, board, bconf)
    return lp, all_projects, lp_users


def create_cards(
        board, bconf, lp_users, lp_projects, filter_bug=None,
        filter_card=None):
    """Create new cards on all projects."""
    if filter_bug or filter_card:
        return
    new_bug_tag = bconf.get("bug_to_card_tag")
    if new_bug_tag is not None:
        for lp_project in lp_projects:
            create_cards_in_project(
                board, lp_project, new_bug_tag, bconf.get("bug_to_card_type"),
                bconf.get("bug_to_card_lane"), lp_users.feature_lanes)


def sync_cards_linked_branches(
        board, bconf, lp, lp_users, filter_bug=None, filter_card=None):
    """Sync any cards that have linked external branches."""
    print "Syncing branch cards"
    for card in board.getCardsWithExternalLinks(only_branches=True):
        if filter_card:
            import ipdb; ipdb.set_trace()
        if card.external_card_id or filter_bug:
            # Cards with external_id are processed by getCardsWithExternalIds
            continue
        branch_name = get_lp_branch_name(card.external_system_url)
        branch = lp.branches.getByUniqueName(unique_name=branch_name)
        if not branch:
            print "Invalid branch url ({}) for card '{}'".format(
                card.external_system_url, card.title)
            continue
        if should_sync_card(card, bconf):
            branch_info = get_branch_info([branch])
            owner_name = BRANCH_MP_REGEX.match(
                card.external_system_url).group("owner")
            card_status = get_card_status(None, '', branch_info)
            card_path = card.lane.path
            if isinstance(card.lane.board, LeankitTaskBoard):
                task_card = card.lane.board.parent_card
                card_path = "{}::{}::{}".format(
                    task_card.lane.path, task_card.title, card_path)
            if branch_info.status and should_move_card(card, bconf):
                assignee = Record(name=owner_name)
                try:
                    move_card(card, card_status, bconf, [assignee], lp_users)
                except IOError as e:
                    print "   * %s (in %s)" % (card.title, card_path)
                    print "     >>> Error moving card:"
                    print "     ", e
                    continue
                kanban_user = lp_users.lp_to_kanban.get(assignee.name, None)
                if kanban_user:
                    card.assigned_user_id = kanban_user.id
                    card.kanban_user = kanban_user
                else:
                    print "   * %s (in %s)" % (card.title, card_path)


def sync_cards_external_ids(
        board, bconf, lp, lp_users, lp_projects, filter_bug=None,
        filter_card=None):
    """Sync any cards that have linked external ids (bugs in launchpad)."""
    print "Syncing cards with external ids"
    for card in board.getCardsWithExternalIds():
        if filter_bug and card.external_card_id != filter_bug:
            continue
        if filter_card:
            import ipdb; ipdb.set_trace()
        if should_sync_card(card, bconf):
            lp_bug = get_lp_bug(card, lp)
            if lp_bug is None:
                continue
            ignore_external_projects = (
                bconf.get("autosync") == "active-projects")
            if ignore_external_projects and not is_bug_in_projects(lp_bug, lp_projects):
                continue  # Skip bugs in external projects
            synced, card_status, assignees, mp_url, mp_type = get_bug_status(
                lp_bug, lp_projects, lp_users.lp_to_kanban.keys())

            card_path = card.lane.path
            if isinstance(card.lane.board, LeankitTaskBoard):
                task_card = card.lane.board.parent_card
                card_path = "{}::{}::{}".format(
                    task_card.lane.path, task_card.title, card_path)

            # Move card to new lane if configured to do so.
            if card_status is not None and should_move_card(card, bconf):
                try:
                    move_card(card, card_status, bconf, assignees, lp_users)
                except IOError as e:
                    print "   * %s: %s (in %s)" % (
                        lp_bug.id, card.title, card_path)
                    print "     >>> Error moving card:"
                    print "     ", e
                    continue
            if synced:
                card.title = lp_bug.title
                for tag in lp_bug.tags:
                    card.addTag(tag)
                if len(assignees) > 0:
                    assignee = assignees[0]
                    kanban_user = lp_users.lp_to_kanban.get(
                        assignee.name, None)
                    if kanban_user:
                        card.assigned_user_id = kanban_user.id
                        card.kanban_user = kanban_user
                if mp_url and mp_type:
                    card.external_system_url = mp_url
                    card.external_system_name = mp_type
                print "   * %s: %s (in %s)" % (lp_bug.id, card.title, card_path)
            else:
                print "   * %s: %s -- Not synced" % (
                    lp_bug.id, card.title)

def move_cards_to_taskboards(
        board, bconf, lp, lp_users, filter_bug=None, filter_card=None):
    """Move branch cards to feature cards if they are landed."""
    if filter_bug:
        return
    if filter_card:
        import ipdb; ipdb.set_trace()
    if "${group}" in bconf["landing_lanes"]:
        # Replace the optional template ${group} with configured board groups.
        # This creates a list of landing_lanes, one per group on the board.
        valid_taskboard_lanes = [
            bconf["landing_lanes"].replace("${group}", group)
            for group in set(lp_users.lp_to_group.values())]
    else:
        valid_taskboard_lanes = bconf["landing_lanes"]

    for feature_card in board.getFeatureCards():
        if (feature_card.lane.path in bconf["deploy_lanes"] or 
            feature_card.lane.path in valid_taskboard_lanes):
            lp_bug = get_lp_bug(feature_card, lp)
            move_cards_to_feature_taskboard(
                feature_card, lp_bug, valid_taskboard_lanes)


def convert_cards_to_work_items(
        board, bconf, lp, lp_users, lp_projects,
        filter_bug=None, filter_card=None):
    """Create work items out of cards if desired."""
    if filter_bug or filter_card:
        # XXX: not yet supported.
        return
    if bconf.get("convert_cards_to_work_items", False) == "on":
        # Sync cards to blueprint work items, if so configured.
        print "   Checking for cards to sync with work items..."
        cards_to_convert = get_blueprint_linked_cards(board)
        work_items = convert_cards_to_work_items(
            cards_to_convert, board, lp_users)
        if len(work_items) == 0:
            print "No cards converted to work items."
        else:
            print "Updating blueprints from work items."
            update_blueprints_from_work_items(lp, work_items)


def move_card(card, bug_status, bconf, assignees, lp_users):
    """Move a card to its right place on the board.

    The right place is considered by looking at the status and the first
    assignee. If the `omnidirectional_card_moves` is turned on in the board
    config, the card can be moved to any lane on the board. If it's not
    turned on, only the lanes that are considered to be "next" of the
    card's current lane are possible move targets.

    If a card is located in a "no_move" lane, the card won't get moved.
    """
    logging.debug("Evaluating for move: {}".format(card.title))
    target_lane_path = bconf[ConfigOptionByStatus[bug_status]]
    if assignees:
        group = lp_users.lp_to_group.get(assignees[0].name)
        if group:
            target_lane_path = target_lane_path.replace("${group}", group)
    omnidirectional_card_moves = (
        bconf.get("omnidirectional_card_moves", "on") == "on")
    no_move_lanes = [
        lane_path.strip()
        for lane_path in bconf.get("no_move_lanes", "").split(",")]
    for no_move_lane_path in no_move_lanes:
        if no_move_lane_path and card.lane.path.endswith(no_move_lane_path):
            logging.debug("Not moving, {} in no_move_lanes".format(
                card.lane.path))
            return
    if omnidirectional_card_moves:
        target_lane = find_card_target_lane_omnidirectional(
            card.lane, target_lane_path)
    else:
        target_lane = find_card_target_lane_next(
            card.lane, target_lane_path)
    logging.debug("    Moving: -> {}".format(target_lane_path))
    card.move(target_lane)


def save_board(board):
    # Split out so we can change the call to a single UpdateCards call.
    for card in board.getCardsWithExternalIds():
        if card.is_dirty:
            if DESCRIPTION_MARKER not in card.description:
                card.description += '\n' + DESCRIPTION_MARKER
                card.description = card.description.strip()

            print "     Saving '%s' for #%s." % (
                "', '".join(card.dirty_attrs), card.external_card_id)
        # Attempt to save the card.  If is it not dirty or new then nothing
        # happens.
        try:
            card.save()
        except IOError as e:
            print "Exception caught.  Skipping board."
            print "  desc:", card.description
            print "  kanban id:", card.assigned_user_id
            print "  LP bug id:", card.external_card_id
            print_exc()
            continue


def main(argv):
    """Update the Yellow team's kanban board with data from Launchpad."""
    args = get_arg_parser().parse_args()
    logging.basicConfig(
        level=logging.DEBUG, format='%(asctime)s %(levelname)s %(message)s')

    if args.debug:
        import httplib
        httplib.HTTPConnection.debuglevel = 1

    config = ConfigParser()
    config.read(args.config)
    global_config, boards_config = parse_config(config)
    user = args.email or global_config['user']
    password = args.password or global_config['password']
    kanban = LeankitKanban(global_config['account'], user, password)

    if args.list:
        print "All active boards:"
        boards = kanban.getBoards()
        for board in boards:
            print " * %s (%d)" % (board.title, board.id)
        return

    if args.board:
        assert args.board in boards_config, (
            "Requested board '%s' is not present in the config file %s." % (
                args.board, args.config))
        board_names = [args.board]
    else:
        board_names = boards_config.keys()
    print "Requested boards:"
    for board_name in board_names:
        print " * %s" % board_name
        try:
            board = kanban.getBoard(title=board_name)
            if board is None:
                print "Could not retrieve board.  Skipping."
                continue
            bconf = boards_config[board_name]
            lp, lp_projects, lp_users = get_lp_data(board, bconf)
            create_cards(
                board, bconf, lp_users, lp_projects, filter_bug=args.bug,
                filter_card=args.card)
            sync_cards_linked_branches(
                board, bconf, lp, lp_users, filter_bug=args.bug,
                filter_card=args.card)
            sync_cards_external_ids(
                board, bconf, lp, lp_users, lp_projects, filter_bug=args.bug,
                filter_card=args.card)
            move_cards_to_taskboards(
                board, bconf, lp, lp_users, filter_bug=args.bug,
                filter_card=args.card)
            convert_cards_to_work_items(
                board, bconf, lp, lp_users, lp_projects, filter_bug=args.bug,
                filter_card=args.card)
            save_board(board)
        except (KeyError, AssertionError) as e:
            print "Exception caught.  Skipping board."
            print_exc()
            continue


if __name__ == '__main__':
    main(sys.argv)
