# Copyright 2012 Canonical Ltd.
"""A script for converting Kanban cards to LP work items."""

import re

from lazr.restfulclient.errors import NotFound
from lp2kanban.kanban import Record


BLUEPRINT_URL_PATTERN = "https://blueprints.launchpad.net/.+/\+spec/.+"

# We map the lane type to the status of the work item. Note that don't
# support POSTPONED work items in this mapping, since the Kanban board
# has no such concept.
LANE_TYPE_TO_WORK_ITEM_STATUS_MAP = {
    1: u"TODO", # Kanban type "Ready"
    2: u"INPROGRESS", # Kanban type "In process"
    3: u"DONE", # Kanban type "Complete"
    }


def convert_card_to_work_item(card, board, lp_users):
    """Convert a single card to a work item string.

    :param card: The card to convert to a work item.
    :param board: The current Kanban board.
    :param lp_users: A `LaunchpadUsersForBoard` instance.
    :return: A `Record` in the form:
        Record({'blueprint': '<blueprint_url>',
                'text': '<work_item_text'})
    """
    blueprint_url_pattern = re.compile(BLUEPRINT_URL_PATTERN)
    blueprint_url = (
        card.description_annotations.get('blueprint_url') or
        card.external_system_url)
    if (blueprint_url is None or not
        blueprint_url_pattern.match(blueprint_url)):
        # Return early if the card isn't linked to a blueprint, since
        # there's no point doing the conversion if we're not going to
        # sync it.
        return None
    if card.description_annotations.get('blueprint_url', None) is None:
        # The card's been linked to the blueprint via it's
        # external_system_url property. This is overwritable with other
        # links, so we store it in the card's description annotations
        # for safe-keeping.
        card.description_annotations.blueprint_url = blueprint_url
        card.save()
    work_item_template = u"{assignee} {title}: {status}"
    assignee_string = ""
    if card.assigned_user_id is not None and card.assigned_user_id != 0:
        board_user = board.users_by_id[card.assigned_user_id]
        lp_user = lp_users.kanban_to_lp.get(board_user)
        if lp_user is not None:
            assignee_string = u"[{lp_user}]".format(lp_user=lp_user.name)
    work_item_status = LANE_TYPE_TO_WORK_ITEM_STATUS_MAP.get(
        card.lane.type, 'TODO')
    work_item_text = work_item_template.format(
        assignee=assignee_string, title=card.title,
        status=work_item_status)
    work_item = Record(
        blueprint_url=blueprint_url, text=work_item_text.strip(),
        status=work_item_status)
    return work_item


def convert_cards_to_work_items(cards, board, lp_users):
    """Convert Kanban cards to blueprint work items.

    :param cards: A sequence of Kanban cards.
    :param board: The current Kanban board.
    :param lp_users: A `LaunchpadUsersForBoard` instance.
    """
    work_item_buckets = {
        'DONE': [],
        'INPROGRESS': [],
        'TODO': [],
        }
    for card in cards:
        work_item = convert_card_to_work_item(card, board, lp_users)
        if work_item is not None:
            # We store the work items by status for now, for the sake of
            # ordering.
            work_item_buckets[work_item.status].append(work_item)
    # re-sort the buckets in alphabetical order; the output of this
    # function should be deterministic.
    for key, bucket in work_item_buckets.items():
        work_item_buckets[key] = sorted(bucket)
    work_items = (
        work_item_buckets['TODO'] + work_item_buckets['INPROGRESS'] +
        work_item_buckets['DONE'])
    return work_items


def fetch_blueprint(lp, url):
    """Fetch and return a Launchpad blueprint."""
    blueprint_url = url.replace(
        "blueprints.launchpad.net", "api.launchpad.net/devel")
    try:
        return lp.load(blueprint_url)
    except NotFound:
        return None


def update_blueprints_from_work_items(lp, work_items):
    """Update blueprints from the work item cards linked to them.

    :param lp: A `launchpadlib.launchpad.Launchpad` instance.
    :param work_items: A sequence of work item Records.
    """
    work_item_text_by_blueprint = {}
    for work_item in work_items:
        if work_item.blueprint_url not in work_item_text_by_blueprint.keys():
            work_item_text_by_blueprint[work_item.blueprint_url] = set()
        work_item_text_by_blueprint[work_item.blueprint_url].add(
            work_item.text)
    for blueprint_url, work_items in work_item_text_by_blueprint.items():
        work_items_text = "\n".join(work_items)
        blueprint = fetch_blueprint(lp, blueprint_url)
        if blueprint is not None:
            print "Adding %s work items to %s" % (
                len(work_items), blueprint.name)
            blueprint.workitems_text = work_items_text
            blueprint.lp_save()


def get_blueprint_linked_cards(board):
    """Return the cards on `board` that need to be converted."""
    blueprint_linked_cards = set()
    for card in board.getCardsWithDescriptionAnnotations():
        blueprint_url = card.description_annotations.get(
            'blueprint_url', None)
        if blueprint_url is not None:
            blueprint_linked_cards.add(card)
    blueprint_url_pattern = re.compile(BLUEPRINT_URL_PATTERN)
    for card in board.getCardsWithExternalLinks():
        if blueprint_url_pattern.match(card.external_system_url):
            blueprint_linked_cards.add(card)
    return blueprint_linked_cards
