# Copyright 2011-2012 Canonical Ltd
"""Tests for the bugs2cards.py script."""

__metaclass__ = type

from ConfigParser import (
    ConfigParser,
    NoSectionError,
    )
from StringIO import StringIO
import tempfile
from textwrap import dedent
import datetime
import unittest

from lp2kanban.bugs2cards import (
    CardStatus,
    create_cards_in_project,
    find_card_target_lane_next,
    get_branch_info,
    is_bug_in_projects,
    get_bug_status,
    get_card_status,
    get_cards_for_feature,
    get_lp_branch_name,
    get_new_bugs,
    LaunchpadUsersForBoard,
    move_card,
    move_cards_to_feature_taskboard,
    parse_config,
    parse_groups_config,
    should_sync_card,
    should_move_card,
    sync_cards_external_ids,
    move_cards_to_taskboards
    )
from lp2kanban.kanban import (
    RESOURCE_TYPE_LINK_SERIES,
    LeankitCard,
    Record,
    )
from lp2kanban.tests.common import (
    FauxLeankitUser,
    FauxLaunchpadUser,
    FauxLaunchpadUsersForBoard,
    FauxBoard,
    FauxCard,
    FauxCardType,
    FauxLane,
    )



def parse_config_text(text):
    config = ConfigParser()
    config.readfp(StringIO(dedent(text)))
    return config


class ConfigReadingTest(unittest.TestCase):

    def test_parse_config_no_globals(self):
        # When 'GLOBAL' section is missing from the configuration,
        # an exception is raised.
        sample_conf = '''
            [Test]
            var = board value
        '''

        config = parse_config_text(sample_conf)

        # NoSectionError is raised because there is no obligatory GLOBAL
        # configuration section.
        self.assertRaises(NoSectionError, parse_config, config)

    def test_parse_config_globals(self):
        # Values in the 'GLOBAL' section are returned separately and
        # are independent from per-board configuration.
        sample_conf = '''
            [GLOBAL]
            var1 = value1
            var2 = value2
            [Board 1]
            var1 = board value 1
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)

        self.assertEqual({'var1': 'value1',
                          'var2': 'value2',
                           }, globals)
        self.assertEqual({'Board 1': {'var1': 'board value 1'}}, boards)

    def test_parse_config_defaults(self):
        # Values from the DEFAULT section are always added to every
        # board configuration section as well.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'board value'}}, boards)

    def test_parse_config_defaults_overridden(self):
        # Each board can override settings from the defaults.
        sample_conf = '''
            [GLOBAL]
            [DEFAULT]
            var = board value
            [Board 1]
            var = my board value
        '''

        config = parse_config_text(sample_conf)

        globals, boards = parse_config(config)
        self.assertEqual({'Board 1': {'var': 'my board value'}}, boards)


class FauxBugBranch:
    def __init__(self, mps, web_link=u''):
        self.branch = Record(landing_targets=mps, web_link=web_link)


class FauxMP:
    def __init__(self, queue_status, target_name, link=None, description=None, date_created=None):
        self.queue_status = queue_status
        self.target_branch = Record(name=target_name)
        self.web_link = link
        self.description = description or ''
        self.date_created = date_created


class FauxLP(object):
    def __init__(self, data={}):
        self.data = data
        self.projects = data.get("projects", {})
        self.project_group = data.get("project_groups", {})
        self.bugs = data.get("bugs", {})
    

class BranchInfoTests(unittest.TestCase):

    def test_bug_with_multiple_branches_based_on_highest_mp_status(self):
        # When a bug has more than one branch associated with it, card status
        # is determined based on the MP with the highest ordered status.
        branches = [FauxBugBranch([FauxMP('Work in progress', 'in-progress')]),
                    FauxBugBranch([FauxMP('Needs review', 'review-branch')])]
        result = get_branch_info(branches)
        self.assertEqual("review-branch", result["target"])

    def test_no_branches_mean_no_info(self):
        # If there are no branches associated with a bug, then no info is
        # returned.
        result = get_branch_info([])
        self.assertEqual(result, {'status': None, 'target': None,
                                  'mp_url': None, 'mp_type': None})

    def test_branches_returns_target(self):
        # The MP's target is pulled out of the bug branch and returned by
        # get_branch_info.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['target'], 'devel')

    def test_branches_returns_mp_link(self):
        # The MP's link is returned as the mp_url.
        branches = [FauxBugBranch([FauxMP('Approved', 'devel',
                                          'http://example.com/mp/link')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'], 'http://example.com/mp/link')

    def test_branches_returns_rietveld_link(self):
        # If, however, a link to https://coderview.appspot.com/\d+ is found
        # in the MP description, then that URL is returned.
        branches = [FauxBugBranch([FauxMP(
            'Approved', 'devel', 'http://example.com/mp/link',
            'https://codereview.appspot.com/1234')])]
        result = get_branch_info(branches)
        self.assertEqual(result['mp_url'],
                         'https://codereview.appspot.com/1234')

    def test_branches_in_review(self):
        # If there is an MP then normally the bug's status is considered "In
        # Review".
        branches = [FauxBugBranch([FauxMP('Fake', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')

    def test_approved_branches_are_approved(self):
        branches = [FauxBugBranch([FauxMP('Approved', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Approved')

    def test_merged_branches_are_merged(self):
        branches = [FauxBugBranch([FauxMP('Merged', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')

    def test_in_progress_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Work in progress', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_rejected_branches_are_in_progress(self):
        branches = [FauxBugBranch([FauxMP('Rejected', 'devel')])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Progress')

    def test_superseded_branches_are_ignored(self):
        branches = [FauxBugBranch([FauxMP('Superseded', 'devel')])]
        result = get_branch_info(branches)
        self.assertIsNone(result["status"])

    def test_highest_ranked_non_final_returned(self):
        # When there are multiple merge proposals and some of them are not in
        # FINAL_STATUSES, return the highest ranked non-final status.
        branches = [FauxBugBranch([FauxMP('Approved', 'approved-branch'),
                                   FauxMP('Rejected', 'bad-branch'),
                                   FauxMP('Needs review', 'review-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')
        self.assertEqual(result['target'], 'review-branch')

    def test_first_highest_ranked_returned(self):
        # If there are multiple merge proposals of the same highest rank, the
        # newest one wins.
        branches = [FauxBugBranch(
            [FauxMP('Work in progress', 'devel'),
             FauxMP('Needs review', 'first'),
             FauxMP('Needs review', 'second'),
             FauxMP('Rejected', 'bad-branch'),
             ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')
        self.assertEqual(result['target'], 'first')

    def test_newest_highest_ranked_returned(self):
        # If there are multiple merge proposals of the same highest rank, but
        # not in final states, the newest one wins.
        now = datetime.datetime.now()
        yesterday = now - datetime.timedelta(days=1)
        lastweek = now - datetime.timedelta(days=7)
        branches = [FauxBugBranch(
            [FauxMP('Needs review', 'lastweek', date_created=lastweek),
             FauxMP('Needs review', 'now', date_created=now),
             FauxMP('Needs review', 'yesterday', date_created=yesterday),
             FauxMP('Rejected', 'bad-branch'),
             ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'In Review')
        self.assertEqual(result['target'], 'now')

    def test_no_inprogress_branches_first_seen_final_state_returned(self):
        # If there are multiple MPs in a final state and no branches in
        # progress, the first final_state branch wins.
        branches = [FauxBugBranch([FauxMP('Merged', 'merged-branch'),
                                   FauxMP('Approved', 'approved-branch'),
                                   ])]
        result = get_branch_info(branches)
        self.assertEqual(result['status'], 'Merged')


class FauxBug:
    def __init__(self, tags=[], title="Faux Bug",
                 description="This is a faux bug.",
                 bug_id=None):
        self.bug_tasks_collection = []
        self.linked_branches = []
        self.tags = tags
        self.title = title
        self.description = description
        self.id = bug_id
        self.saved = False

    def addTask(self, target, status='New', assignee=None):
        record = Record(target=target, status=status, assignee=assignee)
        self.bug_tasks_collection.append(record)
        return record

    def lp_save(self):
        self.saved = True


class FauxTarget:

    def __init__(self, resource_type_link="http://api/project", project=None):
        self.resource_type_link = resource_type_link
        self.project = project
   

class FauxBugTask:

    def __init__(self, bug_id, target='faux-project', tags=None,
                 title="Faux Bug", description="This is a faux bug."):
        if tags is None:
            tags = []
        self.bug = FauxBug(
            bug_id=bug_id, title=title, description=description, tags=tags)
        self.record = self.bug.addTask(target)


class IsBugInProjectsTest(unittest.TestCase):

    def test_false_when_part_of_external_project(self):
        # Return False when the bug is part of an unconfigured (external) project
        bug = FauxBug()
        project = FauxTarget(resource_type_link="project", project="otherproject")
        bug.addTask(project)
        self.assertFalse(is_bug_in_projects(bug, ["myproject1", "myproject2"]))

    def test_true_when_part_of_configured_projects(self):
        # Return True when the bug is one of the configured (internal) projects
        bug = FauxBug()
        project = FauxTarget(resource_type_link="project", project="myproject2")
        bug.addTask(project)
        self.assertTrue(is_bug_in_projects(bug, ["myproject1", "myproject2"]))


    def test_true_when_series_within_a_configured_project(self):
        # Return True when the bug has a series task that is related to one of the
        # configured projects.
        bug = FauxBug()
        series = FauxTarget(
            resource_type_link=RESOURCE_TYPE_LINK_SERIES, project="myproject1")
        bug.addTask(series, status='Fix Released')
        self.assertTrue(is_bug_in_projects(bug, ["myproject1"]))


class BugStatusTest(unittest.TestCase):

    def test_get_bug_status_unknown(self):
        # For a bug on an unrelated project and not assigned to any
        # of the relevant users, it is not considered to match at all.
        bug = FauxBug()
        project = FauxTarget(resource_type_link="project", project="project1")
        bug.addTask(project)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, ['launchpad'], []))
        self.assertFalse(matches)
        self.assertEqual(None, status)
        self.assertEqual([], assignees)

    def test_get_bug_status_assignee(self):
        # For a bug on an unrelated project and assigned to a user
        # we care about, it is returned as matching with proper status.
        bug = FauxBug()
        user = Record(name='user')
        project = FauxTarget(resource_type_link="project", project="project1")
        other_project = FauxTarget(
            resource_type_link="project", project="launchpad")
        bug.addTask(project, status='In Progress', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, [other_project], ['user']))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_project_ignores_non_project_task_status(self):
        # For a bug on an project we care about, it is returned
        # as matching with a status that reflects the project task ignoring
        # non-project tasks.
        bug = FauxBug()
        user = Record(name='user')
        project = FauxTarget(resource_type_link="project", project="project")
        series = FauxTarget(
            resource_type_link=RESOURCE_TYPE_LINK_SERIES, project=project)
        bug.addTask(project, status='In Progress', assignee=user)
        bug.addTask(series, status='Fix Released', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, [project], ['user']))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.CODING, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_multi_targets(self):
        # For a bug with multiple tasks, only some of which are on
        # projects we care about, only assignees from those tasks are
        # returned.
        bug = FauxBug()
        nonuser = Record(name='nonuser')
        user = Record(name='user')
        project = FauxTarget(resource_type_link="project", project="project")
        other_project = FauxTarget(
            resource_type_link="project", project="launchpad")
        bug.addTask(project, status='In Progress', assignee=nonuser)
        bug.addTask(other_project, status='Fix Released', assignee=user)
        matches, status, assignees, mp_url, mp_type = (
            get_bug_status(bug, [project], ["user"]))
        self.assertTrue(matches)
        self.assertEqual(CardStatus.DONE_FIX, status)
        self.assertEqual([user], assignees)

    def test_get_bug_status_multi_statuses(self):
        # For a bug with multiple tasks, all of which are on
        # projects we care about, the bug status is considered to be the
        # "lowest" of all bug task statuses (i.e. 'In Progress' is lower
        # than 'Fix Committed').
        bug = FauxBug()
        project1 = FauxTarget(resource_type_link="project", project="lazr")
        project2 = FauxTarget(resource_type_link="project", project="launchpad")
        bug.addTask(project1, status='Fix Released')
        bug.addTask(project2, status='In Progress')
        matches, status, assignees, mp_url, mp_type = get_bug_status(
            bug, [project1, project2], [])
        self.assertTrue(matches)
        # 'Fix Released' on 'lazr' would put the bug status in DONE state,
        # but since there is an 'In Progress' task as well, it is instead
        # in CODING.
        self.assertEqual(CardStatus.CODING, status)


BRANCH_URL = "https://code.launchpad.net/~me/project/branch-name"
MP_URL = "https://code.launchpad.net/~me/project/branch-name/+merge/1234"
CHARM_BRANCH_URL = ("https://code.launchpad.net/~me/charms/series"
                    "/charm-name/branch")
CHARM_MP_URL = "%s/+merge/288798" % CHARM_BRANCH_URL


class CardStatusTest(unittest.TestCase):

    def test_get_lp_branch_name_branch(self):
        """
        get_lp_branch_name returns a full branch name from the LP branch URL.
        """
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(BRANCH_URL))

    def test_get_lp_branch_name_branch_trailing_slash(self):
        """
        get_lp_branch_name returns a full branch name from the LP URL
        even when the URL ends with a slash.
        """
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(BRANCH_URL + "/"))

    def test_get_lp_branch_name_branch_from_main_host(self):
        """
        Branch name is returned even when the branch URL is on top-level
        launchpad.net (as opposed to code.launchpad.net).
        """
        branch_url = BRANCH_URL.replace("code.", "")
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(branch_url))

    def test_get_lp_branch_name_charm_branch(self):
        """
        get_lp_branch_name returns a full branch name from the LP branch URL
        even for charm branches.
        """
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(CHARM_BRANCH_URL))

    def test_get_lp_branch_name_charm_branch_trailing_slash(self):
        """
        get_lp_branch_name returns a full branch name from the LP URL
        for charm branches even when the URL ends with a slash.
        """
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(CHARM_BRANCH_URL + "/"))

    def test_get_lp_branch_name_charm_branch_from_main_host(self):
        """
        Branch name is returned even when the branch URL is on top-level
        launchpad.net (as opposed to code.launchpad.net).
        """
        branch_url = CHARM_BRANCH_URL.replace("code.", "")
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(branch_url))

    def test_get_lp_branch_name_branch_from_mp(self):
        """
        get_lp_branch_name returns a full branch name from the LP MP URL.
        """
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(MP_URL))

    def test_get_lp_branch_name_branch_from_mp_trailing_slash(self):
        """
        get_lp_branch_name returns a full branch name from the LP MP URL
        even when the URL ends with a slash.
        """
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(MP_URL + "/"))

    def test_get_lp_branch_name_branch_from_mp_on_main_host(self):
        """
        Branch name is returned even when the MP URL is on top-level
        launchpad.net (as opposed to code.launchpad.net).
        """
        mp_url = MP_URL.replace("code.", "")
        self.assertEqual(
            "~me/project/branch-name", get_lp_branch_name(mp_url))

    def test_get_lp_branch_name_from_charm_mp(self):
        """
        get_lp_branch_name returns a full branch name from the LP branch URL
        even for charm branches.
        """
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(CHARM_MP_URL))

    def test_get_lp_branch_name_from_charm_mp_trailing_slash(self):
        """
        get_lp_branch_name returns a full branch name from the LP URL
        for charm branches even when the URL ends with a slash.
        """
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(CHARM_MP_URL + "/"))

    def test_get_lp_branch_name_from_charm_mp_on_main_host(self):
        """
        Branch name is returned even when the branch URL is on top-level
        launchpad.net (as opposed to code.launchpad.net).
        """
        mp_url = CHARM_MP_URL.replace("code.", "")
        self.assertEqual(
            "~me/charms/series/charm-name/branch",
            get_lp_branch_name(mp_url))

    def test_should_sync_card_no(self):
        # Cards with sync markers in neither title nor description are
        # not synced.
        card = Record(title=u'no sync', description=u'really no sync')
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_title_yes(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'sync:', description=u'description')
        self.assertTrue(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_title_no(self):
        # Sync marker for the title is a string 'sync:' at the beginning
        # of the title string, anywhere else and it's ignored.
        card = Record(title=u'no sync:', description=u'description')
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_description_yes(self):
        # Sync marker for the description is a string 'sync:' at the beginning
        # of the title string.
        card = Record(title=u'no sync', description=u'Synced with Launchpad')
        self.assertTrue(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_autosync_no(self):
        # When autosync is 'on', cards with no external card IDs
        # and no branches are not synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=None, external_system_url=None)
        self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_autosync_no_sync_with_non_branch_urls(self):
        # When autosync is 'on', cards with external_system_urls which are not
        # valid launchpad branches are not synced.
        non_branch_urls = [
            "http://www.google.com/",
            "https://bugs.launchpad.net/charms/+source/hacluster/+bug/1",
            "https://code.launchpad.net/~person/project/blah/+merge/111"
        ]
        for url in non_branch_urls:
            card = Record(title=u'no sync', description=u'',
                          external_card_id=None, external_system_url=url)
            self.assertFalse(should_sync_card(card, {'sync_cards': 'on'}))

    def test_should_sync_card_autosync_synced_with_external_card_id(self):
        # When autosync is 'on', cards with external card IDs are synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=u'11', external_system_url=u'')
        self.assertTrue(should_sync_card(card, {'autosync': 'on',
                                                'sync_cards': 'on'}))

    def test_should_sync_card_autosync_synced_with_external_card_id(self):
        # When autosync is 'active-projects', cards with external card IDs are synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=u'11', external_system_url=u'')
        self.assertTrue(should_sync_card(card, {'autosync': 'active-projects',
                                                'sync_cards': 'on'}))

    def test_should_sync_card_autosync_synced_with_branch(self):
        # When autosync is 'on', cards with a valid branch are synced.
        card = Record(title=u'no sync', description=u'',
                      external_card_id=u'11', external_system_url=BRANCH_URL)
        self.assertTrue(should_sync_card(card, {'autosync': 'on',
                                                'sync_cards': 'on'}))

    def test_should_sync_card_autosync_nosync(self):
        # When autosync is 'on', cards with external card IDs or a valid
        # branch are not synced when they contain the no-sync marker.
        card = Record(title=u'(no-sync)', description=u'(no-sync)',
                      external_card_id=u'11', external_system_url=BRANCH_URL)
        self.assertFalse(should_sync_card(card, {'autosync': 'on',
                                                 'sync_cards': 'on'}))

    def test_should_sync_card_no_sync_cards(self):
        # If the sync_cards config option is set to off,
        # should_sync_card() returns False without even looking at the
        # card.
        self.assertFalse(should_sync_card(None, {'sync_cards': 'off'}))

    def test_should_sync_card_no_sync_cards_default(self):
        # If the sync_cards config option isn't specified,
        # should_sync_card() returns False without even looking at the
        # card.
        self.assertFalse(should_sync_card(None, {}))

    def test_no_sync_lanes(self):
        # Cards in a 'no_sync' lane shouldn't get synced.
        board = FauxBoard()
        coding = board.addLane('coding')
        qa = board.addLane('qa')
        conf = {'no_sync_lanes': 'qa', 'autosync': 'on', 'sync_cards': 'on'}
        card1 = coding.addCard(FauxCard(external_card_id='42'))
        card2 = qa.addCard(FauxCard(external_card_id='84'))
        self.assertTrue(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_no_sync_lanes_end(self):
        # Partial lane names can be specified in the 'no_sync_lanes'
        # config, so that every group's QA lane can be set to no_sync,
        # for example.
        board = FauxBoard()
        dev_qa = board.addLane('dev::qa')
        conf = {'no_sync_lanes': 'qa', 'sync_cards': 'on'}
        card = dev_qa.addCard(FauxCard(external_card_id='42'))
        self.assertFalse(should_sync_card(card, conf))

    def test_no_sync_lanes_multiple(self):
        # Multiple 'no_sync' lanes can be separated by comma in the
        # config.
        board = FauxBoard()
        qa = board.addLane('qa')
        review = board.addLane('review')
        conf = {'no_sync_lanes': 'qa,review', 'sync_cards': 'on'}
        card1 = qa.addCard(FauxCard(external_card_id='42'))
        card2 = review.addCard(FauxCard(external_card_id='84'))
        self.assertFalse(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_no_sync_lanes_multiple_strip_whitespace(self):
        # Whitespace is ignored when specifying multiple lanes in the
        # 'no_sync_lanes' config.
        board = FauxBoard()
        qa = board.addLane('qa')
        review = board.addLane('review')
        conf = {'no_sync_lanes': ' qa ,  review ', 'sync_cards': 'on'}
        card1 = qa.addCard(FauxCard(external_card_id='42'))
        card2 = review.addCard(FauxCard(external_card_id='84'))
        self.assertFalse(should_sync_card(card1, conf))
        self.assertFalse(should_sync_card(card2, conf))

    def test_get_card_status_new_no_branch(self):
        # For a bug in 'New', 'Triaged' or 'Confirmed' or 'Incomplete' statuses,
        # the branch status is new state.
        no_branch_info = Record(status=None)
        self.assertEqual(
            CardStatus.NEW,
            get_card_status('New', [], no_branch_info))
        self.assertEqual(
            CardStatus.NEW,
            get_card_status('Confirmed', [], no_branch_info))
        self.assertEqual(
            CardStatus.NEW,
            get_card_status('Triaged', [], no_branch_info))
        self.assertEqual(
            CardStatus.NEW,
            get_card_status('Incomplete', [], no_branch_info))

    def test_get_card_status_coding_no_bug(self):
        # For a card with no associated bug, an 'In Progress' branch status
        # will be in the coding state. 
        branch_info = Record(status='In Progress', target=None)
        self.assertEqual(
            CardStatus.CODING, 
            get_card_status(None, [], branch_info))

    def test_get_card_status_review_no_bug(self):
        # For a card with no associated bug, an 'Approved' or 'In Review'
        # branch status will be in the review state. 
        branch_infos = [
            Record(status='In Review', target=None),
            Record(status='Approved', target=None)]
        for branch_info in branch_infos:
            self.assertEqual(
                CardStatus.REVIEW, 
                get_card_status(None, [], branch_info))

    def test_get_card_status_landing_no_bug(self):
        # For a card with no associated bug, a merged branch status will be in
        # the landing state. 
        branch_info = Record(status='Merged', target=None)
        self.assertEqual(
            CardStatus.LANDING, 
            get_card_status(None, [], branch_info))

    def test_get_card_status_coding_in_progress_no_branch(self):
        # For a bug in 'In Progress' status, it is considered to be in the
        # coding state if there is no branch.
        branch_info = Record(status=None, target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_deploy_fix_committed_no_branch(self):
        # For a bug in 'Fix Committed' status, it is considered to be in the
        # deploy phase when there is no branch linked as the bug is considered
        # fixed with a trivial.
        branch_info = Record(status=None, target=None)
        self.assertEqual(
            CardStatus.DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_coding_branch(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the coding state if there is a branch
        # and it is 'In Progress'.
        branch_info = Record(status='In Progress', target=None)
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.CODING,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_review(self):
        # For a bug in 'In Progress' or 'Fix Committed' status, it is
        # considered to be in the review phase if there is a branch
        # proposed for review.
        branch_info = Record(status='In Review', target=None)
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('In Progress', [], branch_info))
        self.assertEqual(
            CardStatus.REVIEW,
            get_card_status('Fix Committed', [], branch_info))

    def test_get_card_status_landing(self):
        # For a bug in 'In Progress' status, it is considered to be in the
        # landing phase if there is a branch 'Merged'.
        branch_info = Record(status='Merged', target=None)
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_qa(self):
        # For a bug in the 'Fix Committed' status, it is considered
        # to be in the QA phase if there is a branch that has been merged
        # and the bug has a 'qa-needstesting' tag.
        branch_info = Record(status='Merged', target=None)
        self.assertEqual(
            CardStatus.QA,
            get_card_status('Fix Committed', ['qa-needstesting'], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', ['qa-needstesting'], branch_info))

    def test_get_card_status_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'devel'
        branch_info = Record(status='Merged', target='devel')
        self.assertEqual(
            CardStatus.DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_downtime_deploy(self):
        # For a bug in the 'Fix Committed' status, it is
        # considered to be in the downtime-deploy phase if there is a branch
        # that has been merged and the bug has no 'qa-needstesting' tag,
        # and the branch target is 'db-devel'
        branch_info = Record(status='Merged', target='db-devel')
        self.assertEqual(
            CardStatus.DOWNTIME_DEPLOY,
            get_card_status('Fix Committed', [], branch_info))

        # If the bug is still 'In Progress', it is still in the landing phase.
        self.assertEqual(
            CardStatus.LANDING,
            get_card_status('In Progress', [], branch_info))

    def test_get_card_status_done_with_fix(self):
        # For a bug in 'Fix Released' status, it is considered completely done.
        self.assertEqual(
            CardStatus.DONE_FIX,
            get_card_status('Fix Released', [], Record(status=None)))

    def test_get_card_status_done_with_no_fix(self):
        # For a bug in 'Opinion', 'Won't fix', 'Invalid' or'Expired' statuses,
        # it is considered completely done without a fix
        no_branch_info = Record(status=None)
        self.assertEqual(
            CardStatus.DONE_NOFIX,
            get_card_status('Opinion', [], no_branch_info))
        self.assertEqual(
            CardStatus.DONE_NOFIX,
            get_card_status("Won't Fix", [], no_branch_info))
        self.assertEqual(
            CardStatus.DONE_NOFIX,
            get_card_status('Invalid', [], no_branch_info))
        self.assertEqual(
            CardStatus.DONE_NOFIX,
            get_card_status('Expired', [], no_branch_info))


class GetCardsForFeatureTest(unittest.TestCase):

    def setUp(self):
        self.feature_card = FauxCard(external_card_id="123", tags="feature1")
        self.feature_bug = FauxBug(bug_id=123)
        self.feature_bug.linked_branches = [
            FauxBugBranch([], web_link=BRANCH_URL)]
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        self.feature_card.type = feature_type
        self.board = FauxBoard(
            cards=[self.feature_card], feature_cards=[self.feature_card])
        self.feature_card.lane = FauxLane(board=self.board)

    def test_get_cards_for_feature_gets_cards_with_feature_tag(self):
        """Cards sharing any of the tags on a feature card are returned."""
        tagged_card = FauxCard(tags="feature1")
        self.board.cards.extend([tagged_card, FauxCard(tags="otherfeature")])
        cards = get_cards_for_feature(self.feature_card, self.feature_bug)
        self.assertItemsEqual([self.feature_card, tagged_card], cards)

    def test_get_cards_for_feature_gets_cards_with_branches_linked(self):
        """
        Cards whose external_system_url is a linked_branch of the feature bug
        are returned.
        """
        linked_card = FauxCard(
             tags="no-tag-match", external_system_url=BRANCH_URL)
        linked_card_with_trailing_slash = FauxCard(
             tags="no-tag-match", external_system_url=BRANCH_URL+"/")
        mp_linked_card = FauxCard(
             tags="no-tag-match", external_system_url=MP_URL)
        tagged_card = FauxCard(
             tags="feature1",
             external_system_url=BRANCH_URL + "no-branch-match")
        self.board.cards.extend([
            linked_card, tagged_card, mp_linked_card,
            linked_card_with_trailing_slash,
            FauxCard(tags="no-tag-match",
                     external_system_url=BRANCH_URL + "no-branch-match")])
        cards = get_cards_for_feature(self.feature_card, self.feature_bug)
        self.assertItemsEqual([
            self.feature_card, linked_card, linked_card_with_trailing_slash,
            mp_linked_card, tagged_card],
            cards)

    def test_get_cards_for_feature_gets_cards_with_either_tags_or_links(self):
        """
        Cards will be returned if either their tags match a tag on the feature
        card, or the external_system_url is a linked_branch of the feature bug.
        """
        linked_card = FauxCard(external_system_url=BRANCH_URL)
        self.board.cards.extend(
            [linked_card, FauxCard(external_system_url=BRANCH_URL + "-other")])
        cards = get_cards_for_feature(self.feature_card, self.feature_bug)
        self.assertItemsEqual([self.feature_card, linked_card], cards)

    def test_get_cards_for_feature_can_filter_by_card_type(self):
        """Cards not matching card_type will not be returned."""
        branch_card = FauxCard(tags="feature1")
        branch_type = FauxCardType(id=2, name='Branch', is_default=False)
        branch_card.type = branch_type
        self.board.cards.extend([branch_card])
        cards = get_cards_for_feature(
            self.feature_card, self.feature_bug, card_type='Branch')
        self.assertItemsEqual([branch_card], cards)
      

class FindLaneNextTest(unittest.TestCase):

    def test_find_card_target_lane_no_match(self):
        # Lane path needs to end with the string we are looking for.
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)

        self.assertEqual(None, find_card_target_lane_next(coding, 'Landing'))

    def test_find_card_target_lane_exists(self):
        coding = FauxLane('Dev::Coding')
        review = FauxLane('Dev::Review')
        coding.addNextLane(review)
        landing = FauxLane('Dev::Landing')
        review.addNextLane(landing)

        self.assertEqual(landing, find_card_target_lane_next(
            coding, '::Landing'))

    def test_find_card_target_lane_branching(self):
        # All the potential next lanes are searched through recursively
        # for the matching next lane.
        qa = FauxLane('QA::Done')
        land = FauxLane('Land::Quick')
        slow_land = FauxLane('Land::Slow')
        qa.addNextLane(land)
        qa.addNextLane(slow_land)
        rollout = FauxLane('Rollout')
        slow_land.addNextLane(rollout)

        self.assertEqual(rollout, find_card_target_lane_next(qa, 'Rollout'))


class GetNewBugsTest(unittest.TestCase):

    def test_empty_board_no_bugs(self):
        # If the board is empty, and there are no new bugs, no
        # bugs are returned.
        board = FauxBoard(cards=[])
        bugs = []
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_nonempty_board_no_bugs(self):
        # If the board has bugs, but there are no new bugs in the passed
        # in list, no bugs are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = []
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_no_new_bugs(self):
        # If all the passed in bugs are already in the board, no bugs
        # are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = [FauxBug(bug_id=123)]
        self.assertEqual([], list(get_new_bugs(board, bugs)))

    def test_new_bugs(self):
        # The passed in bugs that aren't in the board are returned.
        board = FauxBoard(
            cards=[FauxCard(), FauxCard(external_card_id="123")])
        bugs = [FauxBug(bug_id=123), FauxBug(bug_id=457)]
        self.assertEqual([bugs[1]], list(get_new_bugs(board, bugs)))


class FauxLaunchpadProject:

    def __init__(self, name="MyFauxProject", bug_tasks=None):
        self.name = name
        if bug_tasks is None:
            bug_tasks = []
        self.bug_tasks = bug_tasks

    def searchTasks(self, tags=None):
        bug_tasks = self.bug_tasks
        if tags is not None:
            tags = set(tags)
            bug_tasks = [
                bug_task for bug_task in bug_tasks
                if tags.intersection(bug_task.bug.tags) == tags]
        return bug_tasks


class CreateNewCardsTest(unittest.TestCase):

    def test_no_bugs(self):
        # If no bugs are found, no cards are created.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(bug_tasks=[])
        create_cards_in_project(board, project, 'kanban')
        self.assertEqual([], board.cards)

    def test_new_bugs_card_added(self):
        # If a new bug is found, a card is created for it, having the
        # bug id, title and description saved.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        create_cards_in_project(board, project, 'kanban')
        [card] = board.cards
        self.assertEqual('Test bug', card.title)
        self.assertEqual('Test bug description.', card.description)
        self.assertEqual('42', card.external_card_id)
        self.assertEqual(board.default_cardtype.id, card.type_id)
        self.assertTrue(card.saved)

    def test_new_bugs_tag_removed(self):
        # After a card is created from a bug, the bug tag is removed, so
        # that having a lot of open bugs won't slow down the process.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        create_cards_in_project(board, project, 'kanban')
        [bug_task] = project.bug_tasks
        self.assertEqual([], bug_task.bug.tags)
        self.assertTrue(bug_task.bug.saved)

    def test_new_bugs_tag_pstays_when_save_fails(self):
        # After a card creation from a bug fails, the bug tag is kept.
        board = FauxBoard()
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        old_save = FauxCard.save
        FauxCard.save = lambda x: None
        create_cards_in_project(board, project, 'kanban')
        [bug_task] = project.bug_tasks
        self.assertEqual(['kanban'], bug_task.bug.tags)
        self.assertFalse(bug_task.bug.saved)
        FauxCard.save = old_save

    def test_new_bugs_tagged_only(self):
        # Cards are created only for bugs having the bug tag that is
        # passed in to create_cards_in_project().
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=21, title='Tagged with other tag',
                           tags=['sometag']),
                       FauxBugTask(
                           bug_id=42, title='Tagged with right tag',
                           tags=['kanban']),
                       FauxBugTask(
                           bug_id=42, title='Tagged with no tags',
                           tags=[])])
        create_cards_in_project(board, project, 'kanban')
        [card] = board.cards
        self.assertEqual('Tagged with right tag', card.title)
        self.assertEqual('42', card.external_card_id)

    def test_new_bugs_custom_card_type(self):
        # It's possible to specify which card type that the created card
        # should have, instead of the default one.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        board.cardtypes[feature_type.id] = feature_type
        create_cards_in_project(board, project, 'kanban', cardtype_name='Feature')
        [card] = board.cards
        self.assertEqual(feature_type.id, card.type_id)
        self.assertTrue(card.saved)

    def test_new_bugs_custom_card_lane(self):
        # It's possible to specify to which lane the card should be
        # added.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        bug_lane = board.addLane('incoming')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='Test bug description.',
                           tags=['kanban'])])
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        board.cardtypes[feature_type.id] = feature_type
        create_cards_in_project(
            board, project, 'kanban', cardtype_name='Feature',
            cardlane_path='incoming')
        [card] = board.cards
        self.assertEqual('incoming', card.lane.path)

    def test_bug_with_long_description_truncated(self):
        # Bug descriptions that exceed the 19997 character limit of a
        # kanban card are truncated to 19500 bytes.
        board = FauxBoard()
        bug_lane = board.addLane('bugs')
        project = FauxLaunchpadProject(
            bug_tasks=[FauxBugTask(
                           bug_id=42, title='Test bug',
                           description='x' * 20000,
                           tags=['kanban'])])
        create_cards_in_project(board, project, 'kanban')
        [card] = board.cards
        self.assertEqual('Test bug', card.title)
        self.assertEqual(19500, len(card.description))
        self.assertEqual('42', card.external_card_id)
        self.assertEqual(board.default_cardtype.id, card.type_id)
        self.assertTrue(card.saved)


class GroupsConfigTest(unittest.TestCase):

    def test_parse_group_conf_empty_file(self):
        # If the group config file is empty, parse_groups_config()
        # returns and empty dict.
        sample_conf = ''

        config = parse_config_text(sample_conf)

        self.assertEqual({}, parse_groups_config(config))

    def test_parse_group_conf_no_members(self):
        # If the config file doesn't contain any members,
        # parse_groups_config() returns an empty dict.
        sample_conf = '''
            [Test group]
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual({}, parse_groups_config(config))

    def test_parse_group_conf_single_member(self):
        # If the config file contains a member, parse_groups_config()
        # returns a dict mapping the member to the group.
        sample_conf = '''
            [Test group]
            members = foo
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual({"foo": "Test group"}, parse_groups_config(config))

    def test_parse_group_conf_multiple_members(self):
        # Multiple members in a group are separated by comma.
        sample_conf = '''
            [Test group]
            members = foo,bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group", "bar": "Test group"},
            parse_groups_config(config))

    def test_parse_group_conf_multiple_members_whitespace(self):
        # Whitespace is ignored when separating members with comma.
        sample_conf = '''
            [Test group]
            members =   foo ,  bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group", "bar": "Test group"},
            parse_groups_config(config))

    def test_parse_group_conf_multiple_groups(self):
        # A groups config can contain multiple groups.
        sample_conf = '''
            [Test group 1]
            members = foo
            [Test group 2]
            members = bar
        '''

        config = parse_config_text(sample_conf)

        self.assertEqual(
            {"foo": "Test group 1", "bar": "Test group 2"},
            parse_groups_config(config))



class LaunchpadUsersForBoardTest(unittest.TestCase):

    def setUp(self):
        self.branch_type = FauxCardType(id=1, name='Branch', is_default=False)
        self.feature_type = FauxCardType(
            id=2, name='Feature', is_default=False)

    def test_no_groups_config(self):
        # If the config doesn't contain an entry for a groups config,
        # ane empty dict is created for the group mapping.
        lp = None
        board = FauxBoard(users=[])
        config = {}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)

        self.assertEqual({}, lp_users.lp_to_group)

    def test_with_groups_config(self):
        # If there is an entry for a groups config in the board config,
        # it is read and a groups mapping is created.
        groups_conf = '''
            [Group 1]
            members = foo,bar
            [Group 2]
            members = baz
        '''
        groups_file = tempfile.NamedTemporaryFile()
        groups_file.write(dedent(groups_conf))
        groups_file.flush()

        lp = None
        board = FauxBoard(users=[])
        config = {"groups_config_file": groups_file.name}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)

    def test_feature_lanes_unset_without_feature_cards(self):
        # When no feature cards exist on a board, the feature_lanes dictionary
        # is empty and new_lanes setting is not represented.
        board = FauxBoard(users=[], feature_cards=[])
        coding = board.addLane("dev::coding")
        card = coding.addCard(FauxCard(tags=["sometag"]))
        card.type = self.branch_type
        lp = None
        config = {"new_lanes": "my-Next-lane"}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)
        self.assertEqual({}, lp_users.feature_lanes)

    def test_feature_lanes_unset_without_new_lanes_config(self):
        # When feature cards are present, feature_lanes is empty when no
        # new_lanes setting is in the config.
        feature_card = FauxCard(tags=["sometag"])
        feature_card.type = self.feature_type
        board = FauxBoard(
            users=[], cards=[feature_card], feature_cards=[feature_card])
        coding_lane = board.addLane("dev::coding")
        coding_lane.addCard(feature_card)
        lp = None
        config = {"new_lanes": ""}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)
        self.assertEqual({}, lp_users.feature_lanes)

    def test_feature_lanes_set_with_feature_card_and_static_new_lanes(self):
        # When a feature card is present and tagged and the config contains
        # new_lanes which doesn't specify the ${group} variable, feature_lanes
        # will be set for all tags.
        feature_card = FauxCard(tags="sometag,anothertag")
        feature_card.type = self.feature_type
        board = FauxBoard(
            users=[], cards=[feature_card], feature_cards=[feature_card])
        coding_lane = board.addLane("dev::coding")
        coding_lane.addCard(feature_card)
        lp = None
        config = {"new_lanes": "dev::coding"}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)
        self.assertEqual(
            {"sometag": coding_lane, "anothertag": coding_lane}, 
            lp_users.feature_lanes)

    def test_feature_lanes_unset_with_group_new_lanes_and_no_assignee(self):
        # When a feature card is present and tagged and the new_lanes
        # in the config specifies the ${group} variable, 
        # feature_lanes will be unset when there are no assigned users on the
        # card.
        feature_card = FauxCard(tags="sometag")
        feature_card.type = self.feature_type
        board = FauxBoard(
            users=[], cards=[feature_card], feature_cards=[feature_card])
        coding_lane = board.addLane("dev::group1::coding")
        coding_lane.addCard(feature_card)
        lp = None
        config = {"new_lanes": "dev::${group}::coding"}
        lp_users = LaunchpadUsersForBoard()
        lp_users.set_up_users(lp, board, config)
        self.assertIsNone(feature_card.assigned_user_id)
        self.assertEqual({}, lp_users.feature_lanes)

    def test_feature_lanes_set_for_squad_tags_for_each_configured_group(self):
        # When new_lanes specifies ${group} variable substitution, feature tag
        # lanes will be set for each configured group name.
        groups_conf = '''
            [Group 1]
            members = User1,User2
            [Group 2]
            members = User3
        '''
        groups_file = tempfile.NamedTemporaryFile()
        groups_file.write(dedent(groups_conf))
        groups_file.flush()
        board = FauxBoard(users=[])
        coding_lane = board.addLane("dev::Group 1::coding")
        coding_lane2 = board.addLane("dev::Group 2::coding")
        lp = None
        config = {"groups_config_file": groups_file.name,
                  "new_lanes": "dev::${group}::coding"}
        lp_users = LaunchpadUsersForBoard(kanban_to_lp={1:FauxPerson("User1")})
        lp_users.set_up_users(lp, board, config)
        self.assertEqual(
            {"squad-group-1": coding_lane, "squad-group-2": coding_lane2},
            lp_users.feature_lanes)

    def test_feature_lanes_set_with_group_new_lanes_and_assignee(self):
        # When a feature card is present and tagged and the new_lanes
        # in the config specifies the ${group} variable, 
        # feature_lanes will be set when there are is an assigned user on the
        # card.
        feature_card = FauxCard(tags="sometag,anothertag")
        feature_card.type = self.feature_type
        feature_card.assigned_user_id = 1
        groups_conf = '''
            [Group 1]
            members = User1,User2
            [Group 2]
            members = User3
        '''
        groups_file = tempfile.NamedTemporaryFile()
        groups_file.write(dedent(groups_conf))
        groups_file.flush()
        board = FauxBoard(
            users=[], cards=[feature_card], feature_cards=[feature_card])
        coding_lane = board.addLane("dev::Group 1::coding")
        coding_lane2 = board.addLane("dev::Group 2::coding")
        coding_lane.addCard(feature_card)
        lp = None
        config = {"groups_config_file": groups_file.name,
                  "new_lanes": "dev::${group}::coding"}
        lp_users = LaunchpadUsersForBoard(kanban_to_lp={1:FauxPerson("User1")})
        lp_users.set_up_users(lp, board, config)
        self.assertEqual(
            {"sometag": coding_lane, "anothertag": coding_lane, 
             "squad-group-1": coding_lane, "squad-group-2": coding_lane2}, 
            lp_users.feature_lanes)


class FauxPerson:


    def __init__(self, name):
        self.name = name


class MoveCardsToFeatureTaskboardTest(unittest.TestCase):

    def setUp(self):
        self.feature_card = FauxCard(external_card_id="123", tags="feature1")
        self.feature_bug = FauxBug(bug_id=123)
        self.feature_bug.linked_branches = [
            FauxBugBranch([], web_link=BRANCH_URL)]
        self.branch_type = FauxCardType(id=1, name='Branch', is_default=False)
        self.feature_type = FauxCardType(
            id=2, name='Feature', is_default=False)
        self.feature_card.type = self.feature_type
        self.board = FauxBoard(
            cards=[self.feature_card], feature_cards=[self.feature_card])
        self.coding = self.board.addLane("coding")
        self.landing = self.board.addLane("landing")
        self.deploy = self.board.addLane("deploy")
        self.feature_card.lane = self.coding
        self.lp_users = LaunchpadUsersForBoard()

    def test_move_cards_to_feature_taskboard_ignores_feature_card(self):
        """move_cards_to_feature_taskboard will not move the feature card"""
        move_cards_to_feature_taskboard(
            self.feature_card, self.feature_bug, ["coding"])
        self.assertIsNone(self.feature_card.moved_to)

    def test_move_cards_to_feature_taskboard_ignores_non_source_lanes(self):
        """
        move_cards_to_feature_taskboard will not move the cards that are not in
        one of the provided source_lanes.
        """
        linked_card = FauxCard(external_system_url=BRANCH_URL)
        linked_card.lane = self.coding
        linked_card.type = self.branch_type
        self.board.cards.extend([linked_card])
        move_cards_to_feature_taskboard(
            self.feature_card, self.feature_bug, ["somelane", "otherlane"])
        self.assertIsNone(linked_card.moved_to)

    def test_move_cards_to_feature_taskboard_moves_supported_lanes(self):
        """
        move_cards_to_feature_taskboard moves the cards in any of the
        source_lanes.
        """
        linked_card = FauxCard(external_system_url=BRANCH_URL)
        linked_card.lane = self.landing
        linked_card.type = self.branch_type
        linked_card2 = FauxCard(external_system_url=BRANCH_URL)
        linked_card2.lane = self.deploy
        linked_card2.type = self.branch_type
        self.board.cards.extend([linked_card, linked_card2])
        move_cards_to_feature_taskboard(
            self.feature_card, self.feature_bug, ["landing", "deploy"])
        self.assertEqual(self.feature_card, linked_card.moved_to)
        self.assertEqual(self.feature_card, linked_card2.moved_to)


class MoveCardOmnidirectionalTest(unittest.TestCase):
    """Tests for move_card() with omnidirectional card moves turned on."""

    def setUp(self):
        self.bconf = {
            "omnidirectional_card_moves": "on",
            "coding_lanes": "coding",
            "review_lanes": "review",
            "landing_lanes": "landing",
            "qa_lanes": "qa",
            "deploy_lanes": "deploy",
            "downtime_deploy_lanes": "downtime",
            "done_fix_lanes": "done-with-fix",
            "done_nofix_lanes": "done-without-fix",
        }
        self.board = FauxBoard()
        self.backlog = self.board.addLane("backlog")
        self.board.addLane("coding")
        self.review = self.board.addLane("review")
        self.board.addLane("landing")
        self.qa = self.board.addLane("qa")
        self.board.addLane("deploy")
        self.board.addLane("downtime")
        self.board.addLane("done-with-fix")
        self.board.addLane("done-without-fix")
        self.lp_users = LaunchpadUsersForBoard()

    def test_coding(self):
        # CODING cards are moved to the coding lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual("coding", card.moved_to.path)

    def test_review(self):
        # REVIEW cards are moved to the review lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.REVIEW, self.bconf, [], self.lp_users)
        self.assertEqual("review", card.moved_to.path)

    def test_landing(self):
        # LANDING cards are moved to the landing lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.LANDING, self.bconf, [], self.lp_users)
        self.assertEqual("landing", card.moved_to.path)

    def test_qa(self):
        # QA cards are moved to the qa lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.QA, self.bconf, [], self.lp_users)
        self.assertEqual("qa", card.moved_to.path)

    def test_deploy(self):
        # DEPLOY cards are moved to the deploy lane.
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.DEPLOY, self.bconf, [], self.lp_users)
        self.assertEqual("deploy", card.moved_to.path)

    def test_downtime(self):
        # DOWNTIME_DEPLOY cards are moved to the downtime lane.
        card = self.backlog.addCard(FauxCard())
        move_card(
            card, CardStatus.DOWNTIME_DEPLOY, self.bconf, [], self.lp_users)
        self.assertEqual("downtime", card.moved_to.path)

    def test_done_without_fix(self):
        # DONE_NOFIX cards are moved to the done_nofix_lanes target
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.DONE_NOFIX, self.bconf, [], self.lp_users)
        self.assertEqual("done-without-fix", card.moved_to.path)

    def test_done_with_fix(self):
        # DONE_FIX cards are moved to the done_fix_lanes target
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.DONE_FIX, self.bconf, [], self.lp_users)
        self.assertEqual("done-with-fix", card.moved_to.path)

    def test_backwards(self):
        # Cards can be moved backwards as well, e.g. go from QA to
        # CODING, if an issue is found.
        card = self.qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual("coding", card.moved_to.path)

    def test_group_no_assignee(self):
        # If a target lane requires a group, and the card doesn't have
        # an assignee, the card isn't moved.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_group_assignee_in_group(self):
        # If a target lane requires a group, and the card has an
        # assignee that is part of a group, the card is moved to the
        # group's lane.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        self.lp_users.lp_to_group["foo"] = "test group"
        move_card(
            card, CardStatus.CODING, self.bconf, [FauxPerson("foo")],
            self.lp_users)
        self.assertEqual("test group::coding", card.moved_to.path)

    def test_group_assignee_not_in_group(self):
        # If a target lane requires a group, and the card has an
        # assignee that is not part of a group, the card isn't moved.
        self.bconf["coding_lanes"] = "${group}::coding"
        self.board.addLane("test group::coding")
        card = self.backlog.addCard(FauxCard())
        self.lp_users.lp_to_group["foo"] = "test group"
        move_card(
            card, CardStatus.CODING, self.bconf, [FauxPerson("bar")],
            self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes(self):
        # Cards in a "no_move" lane doesn't get moved by move_card().
        self.bconf["no_move_lanes"] = "qa"
        card = self.qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes_end(self):
        # Partial lane names can be specified in the "no_move_lanes"
        # config, so that every group's QA lane can be set to no_move,
        # for example.
        dev_qa = self.board.addLane("dev::qa")
        self.bconf["no_move_lanes"] = "qa"
        card = dev_qa.addCard(FauxCard())
        move_card(card, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card.moved_to)

    def test_no_move_lanes_multiple(self):
        # Multiple "no_move" lanes can be separated by comma in the
        # config.
        self.bconf["no_move_lanes"] = "qa,review"
        card1 = self.qa.addCard(FauxCard())
        card2 = self.review.addCard(FauxCard())
        move_card(card1, CardStatus.CODING, self.bconf, [], self.lp_users)
        move_card(card2, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card1.moved_to)
        self.assertEqual(None, card2.moved_to)

    def test_no_move_lanes_multiple_strip_whitespace(self):
        # Whitespace is ignored when specifying multiple lanes in the
        # "no_move_lanes" config.
        self.bconf["no_move_lanes"] = " qa ,  review "
        card1 = self.qa.addCard(FauxCard())
        card2 = self.review.addCard(FauxCard())
        move_card(card1, CardStatus.CODING, self.bconf, [], self.lp_users)
        move_card(card2, CardStatus.CODING, self.bconf, [], self.lp_users)
        self.assertEqual(None, card1.moved_to)
        self.assertEqual(None, card2.moved_to)


class MoveCardTest(unittest.TestCase):

    def setUp(self):
        self.branch_type = FauxCardType(
            id=1, name='Branch', is_default=False)

    def test_should_move_not_if_disabled(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        card.type = self.branch_type
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {'move_cards': 'off'}))

    def test_should_move_not_if_disabled_and_the_default_is_disabled(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        card.type = self.branch_type
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {}))

    def test_should_move_not_if_archived(self):
        board = FauxBoard(cards=[], is_archived=True)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        card.type = self.branch_type
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {'move_cards': 'on'}))

    def test_should_move_if_enabled_and_not_archived(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        card.type = self.branch_type
        board.cards.append(card)
        self.assertTrue(should_move_card(card, {'move_cards': 'on'}))

    def test_should_move_not_if_feature_card_type(self):
        board = FauxBoard(cards=[], is_archived=False)
        lane = FauxLane(board=board)
        card = FauxCard(lane=lane)
        feature_type = FauxCardType(id=2, name='Feature', is_default=False)
        card.type = feature_type
        board.cards.append(card)
        self.assertFalse(should_move_card(card, {'move_cards': 'on'}))

class SyncBoardTest(unittest.TestCase):

    def setUp(self):
        """Create a board with a feature feature card with tag feature1"""
        feature_type = FauxCardType(id=2, name='Feature1', is_default=False)
        self.branch_type = FauxCardType(id=1, name='Branch', is_default=False)
        self.feature_card = FauxCard(tags='feature1', title='Feature1')
        self.feature_card.type = feature_type
        self.board = FauxBoard(
            cards=[self.feature_card], feature_cards=[self.feature_card])
        self.bconf = {
            'landing_lanes': 'landing',
            'deploy_lanes': 'deploy',
        }
        self.coding_lane = FauxLane(board=self.board, path='coding')
        self.landing_lane = FauxLane(board=self.board, path='landing')
        self.deploy_lane = FauxLane(board=self.board, path='deploy')
        self.lp_users = FauxLaunchpadUsersForBoard()

    def test_sync_cards_moves_landing_cards_to_taskboard_in_landing_lane(self):
    	"""
        Feature-linked cards in landing lanes will move to taskboard when the
        feature card is in landing_lanes.
        """
        self.feature_card.lane = self.landing_lane
        coding_card = FauxCard(tags='feature1', title='CodingCard')
        coding_card.type = self.branch_type
        coding_card.lane = self.coding_lane
        landed_card = FauxCard(tags='feature1', title='LandedCard')
        landed_card.type = self.branch_type
        landed_card.lane = self.landing_lane
        self.board.cards.extend([coding_card, landed_card])
        move_cards_to_taskboards(
            self.board, self.bconf, lp=FauxLP(), lp_users=self.lp_users)
        self.assertIsNone(coding_card.moved_to)
        self.assertEqual(self.feature_card, landed_card.moved_to)

    def test_sync_cards_moves_landing_cards_to_taskboard_in_deploy_lane(self):
    	"""
        Feature-linked cards in landing lanes will move to taskboard when the
        feature card is in deploy_lanes.
        """
        self.feature_card.lane = self.deploy_lane
        coding_card = FauxCard(tags='feature1', title='CodingCard')
        coding_card.type = self.branch_type
        coding_card.lane = self.coding_lane
        landed_card = FauxCard(tags='feature1', title='LandedCard')
        landed_card.type = self.branch_type
        landed_card.lane = self.landing_lane
        self.board.cards.extend([coding_card, landed_card])
        move_cards_to_taskboards(
            self.board, self.bconf, lp=FauxLP(), lp_users=self.lp_users)
        self.assertIsNone(coding_card.moved_to)
        self.assertEqual(self.feature_card, landed_card.moved_to)

    def test_sync_cards_moves_no_cards_to_taskboard_not_landed_or_deploy(self):
    	"""
        Feature-linked cards in landing lanes will not move to taskboard when
        the feature card is not in the landing_lanes or deploy_lanes.
        """
        self.feature_card.lane = self.coding_lane
        landed_card = FauxCard(tags='feature1', title='LandedCard')
        landed_card.type = self.branch_type
        landed_card.lane = self.landing_lane
        self.board.cards.extend([landed_card])
        move_cards_to_taskboards(
            self.board, self.bconf, lp=FauxLP(), lp_users=self.lp_users)
        self.assertIsNone(landed_card.moved_to)

    def test_sync_cards_moves_landing_cards_to_taskboard_with_groups(self):
    	"""
        When landing_lanes contains ${groups} template, feature-linkd cards in
        all configured  landing lanes will be moved to the feature taskboard.
        """
        self.bconf = {
            'landing_lanes': '${group}::landing',
            'deploy_lanes': 'deploy',
        }
        landing_lane1 = FauxLane(board=self.board, path='ALPHA::landing')
        landing_lane2 = FauxLane(board=self.board, path='BETA::landing')
        lp_users = FauxLaunchpadUsersForBoard(
            lp_to_group={"person1": "ALPHA", "person2": "BETA"})
        self.feature_card.lane = self.deploy_lane
        coding_card = FauxCard(tags='feature1', title='CodingCard')
        coding_card.type = self.branch_type
        coding_card.lane = self.coding_lane
        landed_card1 = FauxCard(tags='feature1', title='AlphaLandedCard')
        landed_card1.type = self.branch_type
        landed_card1.lane = landing_lane1
        landed_card2 = FauxCard(tags='feature1', title='BetaLandedCard')
        landed_card2.type = self.branch_type
        landed_card2.lane = landing_lane2
        self.board.cards.extend([coding_card, landed_card1, landed_card2])
        move_cards_to_taskboards(
            self.board, self.bconf, lp=FauxLP(), lp_users=lp_users)
        self.assertIsNone(coding_card.moved_to)
        self.assertEqual(self.feature_card, landed_card1.moved_to)
        self.assertEqual(self.feature_card, landed_card2.moved_to)

    def test_sync_cards_skips_bugs_of_external_projects_when_active_projects_set(self):
    	"""Bugs of external projects are skipped when autosync set to active-projects."""
        # Configure bugs2cards to sync only active-projects myproject1 and myproject2
        self.bconf.update({"autosync": "active-projects",
                           "projects": "myproject,myproject2",
                           "sync_cards": "on", "move_cards": "on",
                           "new_lanes": "new"})
        # Setup bug1 and bug2 in separate projects which both need to move coding -> new
        project1 = FauxTarget(resource_type_link="project", project="myproject")
        project2 = FauxTarget(resource_type_link="project2", project="myproject2")
        otherproject = FauxTarget(resource_type_link="project", project="otherproject")
        lp_users = FauxLaunchpadUsersForBoard(lp_to_kanban={"person1": "Person One"})
        lp_projects = [project1, project2]
        bug_task1 = FauxBugTask(
            bug_id=123, target=project1, title='Test bug in myproject',
            description='Test bug description.')
        bug_task2 = FauxBugTask(
            bug_id=456, target=otherproject, title='Test bug in otherproject',
            description='Test bug description.')
        bug_card1 = FauxCard(
            title='BugCard123', external_card_id=u'123')
        bug_card2 = FauxCard(
            title='BugCard456', external_card_id=u'456')
        board = FauxBoard(
            cards=[bug_card1, bug_card2], bug_cards=[bug_card1, bug_card2])
        new_lane = board.addLane('new')
        coding_lane = board.addLane('coding')
        bug_card1.lane = coding_lane
        bug_card2.lane = coding_lane
        bug_type = FauxCardType(id=3, name='Bug', is_default=False)
        bug_card1.type = bug_type
        bug_card2.type = bug_type
        self.board.cards.extend([bug_card1, bug_card2])
        # Include project1 and project2 in our active 'projects' fake LP data
        lp = FauxLP({"projects": {"myproject": project1, "myproject2": project2},
                 "bugs": {123: bug_task1.bug, "456": bug_task2.bug}})
        sync_cards_external_ids(board, self.bconf, lp, lp_users, lp_projects)
        coding_lane = FauxLane(board=board, path='coding')
        self.assertEqual(new_lane, bug_card1.moved_to)
        self.assertIsNone(bug_card2.moved_to)
