# Copyright 2012 Canonical Ltd.
__metaclass__ = type

import re
import testtools

from lp2kanban import cards2workitems
from lp2kanban.cards2workitems import (
    LANE_TYPE_TO_WORK_ITEM_STATUS_MAP,
    convert_card_to_work_item,
    convert_cards_to_work_items,
    get_blueprint_linked_cards,
    update_blueprints_from_work_items,
    )
from lp2kanban.kanban import (
    Record,
    )
from lp2kanban.tests.common import (
    FauxBoard,
    FauxCard,
    FauxLane,
    FauxLeankitUser,
    FauxLaunchpadUser,
    FauxLaunchpadUsersForBoard,
    )
from random import shuffle
from testtools.monkey import patch
from textwrap import dedent


class CardsToWorkItemsBase(testtools.TestCase):

    def setUp(self):
        super(CardsToWorkItemsBase, self).setUp()
        self.user_1 = FauxLeankitUser(
            id=100, user_name="test@canonical.com", full_name="Some Dev",
            email_address="test@canonical.com")
        self.lp_user_1 = FauxLaunchpadUser(name='foobar')
        kanban_to_lp = {
            self.user_1: self.lp_user_1,
            }
        self.lp_users_for_board = FauxLaunchpadUsersForBoard(
            kanban_to_lp=kanban_to_lp)
        self.board = FauxBoard(
            cards=[], users_by_id={self.user_1.id: self.user_1})
        self.blueprint_url = "https://blueprints.launchpad.net/foo/+spec/bar"
        self.default_lane = FauxLane(type=1)

    def _assertMatchesWorkItemFormat(self, string):
        """Assert that a string matches the work item format."""
        pattern = (
            "(\[.*\]\s+)?(.+)\s*:\s*([A-Z]+)")
        match = re.match(pattern, string)
        self.assertTrue(
            match is not None,
            "String \"{string}\" did not match pattern \"{pattern}\"".format(
                string=string, pattern=pattern))

    def _generateCard(self, board=None, lane=None,
                     description_annotations=None, assigned_user_id=0,
                     title=None, external_system_url=None):
        if board is None:
            board = self.board
        if lane is None:
            lane = self.default_lane
        if description_annotations is None:
            description_annotations = Record(
                blueprint_url=self.blueprint_url)
        if title is None:
            title = "A card"
        card = FauxCard(
            title=title, lane=lane, assigned_user_id=assigned_user_id,
            description_annotations=description_annotations,
            external_system_url=external_system_url)
        board.cards.append(card)
        return card


class CardsToWorkItemsTestCase(CardsToWorkItemsBase):
    """Tests for the cards2workitems.py script."""

    def setUp(self):
        super(CardsToWorkItemsTestCase, self).setUp()

    def test_returns_equal_numbers(self):
        # convert_cards_to_work_items() takes a sequence of Kanban cards
        # and returns one work item per card.
        cards = [self._generateCard(title="Card %i" % id) for id in range(5)]
        work_items = convert_cards_to_work_items(
            cards, self.board, self.lp_users_for_board)
        self.assertEqual(len(cards), len(work_items))

    def test_returns_correct_format(self):
        # Each item returned by convert_cards_to_work_items() is of the
        # correct work item format.
        cards = [
            self._generateCard(
                title="Test card 1", lane=FauxLane(type=1)),
            self._generateCard(
                title="Test card 2",
                assigned_user_id=self.user_1.id,
                lane=FauxLane(type=2)),
            self._generateCard(
                title="Test card 3", lane=FauxLane(type=3)),
            ]
        work_items = convert_cards_to_work_items(
            cards, self.board, self.lp_users_for_board)
        for work_item in work_items:
            self._assertMatchesWorkItemFormat(work_item.text)

    def test_unlinked_cards_ignored(self):
        # If a card is not linked to a blueprint it will not be
        # included in the set() of work items returned by
        # convert_cards_to_work_items()
        cards = [self._generateCard(title="Card %i" % id) for id in range(3)]
        cards.append(self._generateCard(
            description_annotations=Record(blueprint_url='http://nyan.cat')))
        work_items = convert_cards_to_work_items(
            cards, self.board, self.lp_users_for_board)
        self.assertEqual(len(cards) - 1, len(work_items))
        for item in work_items:
            self.assertIsNotNone(item)

    def tests_workitems_ordered_by_status(self):
        # convert_cards_to_work_items() returns work items ordered by
        # status: TODO, INPROGRESS, DONE.
        cards = [
            self._generateCard(
                title="Test card 1", lane=FauxLane(type=1)),
            self._generateCard(
                title="Test card 2",
                lane=FauxLane(type=2)),
            self._generateCard(
                title="Test card 3", lane=FauxLane(type=3)),
            self._generateCard(
                title="Test card 4", lane=FauxLane(type=1)),
            self._generateCard(
                title="Test card 5",
                lane=FauxLane(type=2)),
            self._generateCard(
                title="Test card 6", lane=FauxLane(type=3)),
            ]
        shuffle(cards)
        work_items = convert_cards_to_work_items(
            cards, self.board, self.lp_users_for_board)
        actual_work_items = "\n".join(
            [item.text for item in work_items])
        expected_work_items = dedent("""
            Test card 1: TODO
            Test card 4: TODO
            Test card 2: INPROGRESS
            Test card 5: INPROGRESS
            Test card 3: DONE
            Test card 6: DONE
            """).strip()
        self.assertEqual(expected_work_items, actual_work_items)


class CardToWorkItemTestCase(CardsToWorkItemsBase):
    """Tests for cards2workitems.convert_card_to_work_item()"""

    def setUp(self):
        super(CardToWorkItemTestCase, self).setUp()

    def test_converts_title(self):
        # The description of a work item is made up of the title of the
        # card.
        card = self._generateCard()
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        title, status = work_item.text.split(":")
        self.assertEqual(title.strip(), card.title)

    def test_converts_lane_type_to_status(self):
        # The type of the lane in which the card resides is converted
        # into a work item status by convert_card_to_work_item().
        card = self._generateCard()
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        title, status = work_item.text.split(":")
        self.assertEqual(
            LANE_TYPE_TO_WORK_ITEM_STATUS_MAP[card.lane.type],
            status.strip())

    def unknown_lane_types_convert_to_TODO(self):
        # If the lane type isn't one that's in
        # LANE_TYPE_TO_WORK_ITEM_STATUS_MAP, convert_card_to_work_item()
        # will convert the status to TODO by default.
        card = self._generateCard(lane=FauxLane(type=999))
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        title, status = work_item.text.split(":")
        self.assertEqual("TODO", status.strip())

    def test_converts_assignee(self):
        # convert_card_to_work_item() converts the assigned user ID of
        # the card and converts it to a Launchpad user.
        card = self._generateCard(assigned_user_id=self.user_1.id)
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        match = re.match(
            "\[(?P<assignee>.*)\]", work_item.text)
        assignee = match.groups()[0]
        self.assertEqual(self.lp_user_1.name, assignee)

    def test_matches_workitem_format(self):
        # convert_card_to_work_item() returns a string formatted to
        # match the Launchpad work item format of:
        #   [$assignee] $title: $STATUS
        card = self._generateCard()
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        self._assertMatchesWorkItemFormat(work_item.text)

    def test_blueprint_extracted_from_card_description_annotations(self):
        # The blueprint onto which to sync a card's work item can be
        # found by inspecting the card's description annotations. This
        # is returned as work_item.blueprint_url.
        card = self._generateCard()
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        self.assertEqual(
            card.description_annotations.blueprint_url, work_item.blueprint_url)

    def test_blueprint_extracted_from_card_external_link(self):
        # Cards can also be linked to blueprints by putting the
        # blueprint url in the card's external_system_url property
        # (via the web UI). If convert_card_to_work_item() encounters
        # this it will also store the blueprint URL in the card's
        # description annotations so that it's less likely to be
        # overwritten.
        card = self._generateCard(
            external_system_url=self.blueprint_url,
            description_annotations=Record(),
            )
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        self.assertEqual(
            self.blueprint_url, card.description_annotations.blueprint_url)

    def test_non_blueprint_links_ignored(self):
        # If the card has some other external link attached to it that
        # isn't a blueprint, convert_card_to_work_item() will return
        # None.
        card = self._generateCard(
            description_annotations=Record
                (blueprint_url= 'https://nyancat.com'))
        work_item = convert_card_to_work_item(
            card, self.board, self.lp_users_for_board)
        self.assertIsNone(work_item)


class FauxBlueprint(Record):

    def lp_save(self):
        """A stub that does nothing, just for compatibility's sake."""
        self.saved = True


class UpdateBluePrintsFromWorkItemsTestCase(CardsToWorkItemsBase):
    """Tests for update_blueprints_from_work_items()."""

    def setUp(self):
        super(UpdateBluePrintsFromWorkItemsTestCase, self).setUp()
        blueprint_url_tpl = (
            "https://blueprints.launchpad.net/foo/+spec/bar%i")
        self.cards = [
            self._generateCard(
                title="Card %i" % id,
                description_annotations=Record(
                    blueprint_url=blueprint_url_tpl % id))
                for id in range(3)]
        self.work_items = convert_cards_to_work_items(
            self.cards, self.board, self.lp_users_for_board)
        self.blueprints = {}
        for work_item in self.work_items:
            self.blueprints[work_item.blueprint_url] = FauxBlueprint(
                workitems_text='', saved=False, name="blueprint")

    def test_fetches_blueprints(self):
        # update_blueprints_from_work_items() fetches the blueprints frm
        # Launchpad.
        # We stub out the Launchpad interaction using a fake version of
        # fetch_blueprint()
        called_with_urls = []
        def fake_fetch_blueprint(lp, url):
            called_with_urls.append(url)
            return FauxBlueprint(name="somename")
        self.addCleanup(
            patch(cards2workitems, "fetch_blueprint", fake_fetch_blueprint))
        update_blueprints_from_work_items(lp=None, work_items=self.work_items)
        blueprint_urls = [item.blueprint_url for item in self.work_items]
        self.assertEqual(sorted(blueprint_urls), sorted(called_with_urls))

    def test_updates_blueprint_work_items(self):
        # update_blueprints_from_work_items() updates the workitems_text
        # attribute of each blueprint.
        def fake_fetch_blueprint(lp, url):
            return self.blueprints[url]
        self.addCleanup(
            patch(cards2workitems, "fetch_blueprint", fake_fetch_blueprint))
        update_blueprints_from_work_items(lp=None, work_items=self.work_items)
        for blueprint in self.blueprints.values():
            self.assertNotEqual('', blueprint.workitems_text)

    def test_saves_blueprint(self):
        # update_blueprints_from_work_items() calls lp_save() on the
        # blueprint.
        def fake_fetch_blueprint(lp, url):
            return self.blueprints[url]
        self.addCleanup(
            patch(cards2workitems, "fetch_blueprint", fake_fetch_blueprint))
        update_blueprints_from_work_items(lp=None, work_items=self.work_items)
        for blueprint in self.blueprints.values():
            self.assertTrue(blueprint.saved)

    def test_ignores_nonexistent_blueprints(self):
        # If update_blueprints_from_work_items() encounters a
        # nonexistent blueprint (say from a bad link), it simply ignores
        # it.
        def fake_fetch_blueprint(lp, url):
            return None
        self.addCleanup(
            patch(cards2workitems, "fetch_blueprint", fake_fetch_blueprint))
        update_blueprints_from_work_items(lp=None, work_items=self.work_items)
        # No assert here; we just want to make sure that
        # update_blueprints_from_work_items() doesn't error.


class GetBluePrintLinkedCardsTestCase(CardsToWorkItemsBase):
    """Tests for get_blueprint_linked_cards()."""

    def test_returns_cards_with_blueprint_url_annotations(self):
        # get_blueprint_linked_cards() returns cards with a blueprint url in
        # their description_annotations.
        blueprint_card = self._generateCard(
            title="Card 1",
            description_annotations=Record(blueprint_url=self.blueprint_url))
        self.board._cards_with_description_annotations.add(blueprint_card)
        cards_to_convert = get_blueprint_linked_cards(self.board)
        self.assertIn(blueprint_card, cards_to_convert)

    def test_doesnt_return_non_blueprint_cards(self):
        # get_blueprint_linked_cards() doesn't return cards that don't have a
        # blueprint_url set.
        non_blueprint_card = self._generateCard(
            title="Card 1", description_annotations=Record(foo='bar'))
        self.board._cards_with_description_annotations.add(non_blueprint_card)
        cards_to_convert = get_blueprint_linked_cards(self.board)
        self.assertNotIn(non_blueprint_card, cards_to_convert)

    def test_returns_cards_with_external_link_to_blueprint(self):
        # get_blueprint_linked_cards() returns cards that have an
        # external_system_url that points to a blueprint.
        blueprint_card = self._generateCard(
            title="Card 1", description_annotations={})
        blueprint_card.external_system_url = self.blueprint_url
        self.board._cards_with_external_links.add(blueprint_card)
        cards_to_convert = get_blueprint_linked_cards(self.board)
        self.assertIn(blueprint_card, cards_to_convert)

    def test_does_not_return_cards_with_links_not_to_blueprints(self):
        # get_blueprint_linked_cards() won't return cards that have a
        # external_system_url that points to something other than a
        # blueoprint.
        non_blueprint_card = self._generateCard(
            title="Card 1", description_annotations={})
        non_blueprint_card.external_system_url = "http://example.com"
        self.board._cards_with_external_links.add(non_blueprint_card)
        cards_to_convert = get_blueprint_linked_cards(self.board)
        self.assertNotIn(non_blueprint_card, cards_to_convert)
