# Copyright 2011-2012 Canonical Ltd
"""Tests for Kanban card description annotations."""

from lp2kanban.kanban import (
    LeankitCard,
    Record,
    )
from lp2kanban.tests.common import FauxLane
from textwrap import dedent
import unittest

class DescriptionAnnotationsTest(unittest.TestCase):
    """Tests for kanban.LeankitCard description annotations."""

    def test_parsed_description_parses_json_in_description(self):
        # parse_card_description() parses key:value pairs out of a
        # card's description and returns them as a dict.
        card = LeankitCard.create(FauxLane("Test"))
        card.title=u"some card"
        card.description = dedent("""
        {"key2": "value2", "key1": "value 1"}
        """)
        parsed_description, _1, _2 = card.parsed_description
        self.assertEqual("value 1", parsed_description['key1'])
        self.assertEqual("value2", parsed_description['key2'])

    def test_parsed_description_returns_other_text_unparsed(self):
        # Any text not in the key=value form is returned unparsed by
        # card.parsed_description().
        card = LeankitCard.create(FauxLane("Test"))
        card.title=u"some card"
        card.description = dedent("""
            Some other text,
            {"key2": "value2", "key1": "value 1"}
            And yet some more text besides.
            """)
        _, text_before_json, text_after_json = card.parsed_description
        self.assertEqual(
            "Some other text,", text_before_json)
        self.assertEqual(
            "And yet some more text besides.", text_after_json)

    def test_description_annotations_returns_only_key_value_pairs(self):
        # LeankitCard.description_annotations returns only the key=value
        # pairs from the description field.
        card = LeankitCard.create(FauxLane("Test"))
        card.title=u"some card"
        card.description = dedent("""
            {"key2": "value2", "key1": "value 1"}
            Here's a URL: http://test.com?a=b
            """)
        self.assertEqual(
            {'key1': 'value 1', 'key2': 'value2'},
            card.description_annotations)

    def test_setting_description_annotations_overwrites_annotations(self):
        # LeankitCard._setDescriptionAnnotations() can be used to write
        # new annotations to the description field.
        card = LeankitCard.create(FauxLane("Test"))
        card.title=u"some card"
        card.description = dedent("""
            Some text
            {"key2": "value2", "key1": "value 1"}
            """)
        card._setDescriptionAnnotations({'spam': 'eggs', 'ham': 'spam'})
        expected_description = dedent("""
            Some text
            {"ham": "spam", "spam": "eggs"}
            """).strip()
        self.assertEqual(expected_description, card.description.strip())

    def test_annotations_is_a_record(self):
        # LeankitCard.description_annotations is a Record, so it hass
        # attribute access for its values.
        card = LeankitCard.create(FauxLane("Test"))
        card.title=u"some card"
        card.description = '{"key1": "value 1"}'
        self.assertIsInstance(card.description_annotations, Record)
        self.assertEqual("value 1", card.description_annotations.key1)

    def test_empty_annotations_returns_empty_record(self):
        # If there are no annotations on a card, description_annotations
        # will simply return an empty record.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = "Hello, world!"
        self.assertEqual(Record(), card.description_annotations)

    def test_adding_annotations_to_text_only_description_works(self):
        # If there's already a description, _setDescriptionAnnotations()
        # will add annotations after the existing description.
        card = LeankitCard.create(FauxLane("Test"))
        new_annotations = {
            'var1': 'value 1',
            'var2': 'value 2',
            }
        card.description = "Hello, world!"
        card._setDescriptionAnnotations(new_annotations)
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual("Hello, world!", text_before)
        self.assertEqual("", text_after)
        self.assertEqual(new_annotations, annotations)

    def test_parsing_annotations_handles_native_objects(self):
        # If there are objects embedded in the JSON-encoded annotations,
        # they will be loaded into the description_annotations Record
        # correctly.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{"foo": {"bar": true}}\nxyz\n'
        annotations = card.description_annotations
        self.assertEqual({u'foo': {u'bar': True}}, annotations)

    def test_parsing_annotations_handles_no_text_before(self):
        # parsed_description correctly handles situations where there is
        # no text before the JSON.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = '{"foo": "bar"}\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('', text_before)
        self.assertEqual('xyz', text_after.strip())

    def test_parsing_annotations_handles_no_text_after(self):
        # parsed_description correctly handles situations where there is
        # no text after the JSON.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'xyz\n{"foo": "bar"}\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('', text_after)
        self.assertEqual('xyz', text_before.strip())

    def test_parsing_annotations_handles_multiple_lines_before(self):
        # parsed_description correctly handles situations where there
        # are multiple lines of text before the JSON.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\ndef \n{"foo": "bar"}\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('abc\ndef', text_before.strip())
        self.assertEqual('xyz', text_after)

    def test_parsing_annotations_handles_multiple_lines_after(self):
        # parsed_description correctly handles situations where there
        # are multiple lines of text after the JSON.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{"foo": "bar"}\nuvw\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('abc', text_before.strip())
        self.assertEqual('uvw\nxyz', text_after)

    def test_parsing_annotations_handles_whitespace_before_brace(self):
        # parsed_description correctly handles situations where there
        # is whitespace before the opening brace of the JSON
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{"foo": "bar"}\nuvw\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('abc', text_before.strip())
        self.assertEqual('uvw\nxyz', text_after)

    def test_parsing_annotations_handles_whitespace_after_brace(self):
        # parsed_description correctly handles situations where there
        # is whitespace after the closing brace of the JSON
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{"foo": "bar"}    \nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('abc', text_before)
        self.assertEqual('xyz', text_after)

    def test_parsing_annotations_handles_no_json(self):
        # parsed_description correctly handles situations where there
        # there is no JSON in the description.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({}, annotations)
        self.assertEqual('abc\nxyz', text_before.strip())
        self.assertEqual('', text_after)

    def test_parsing_annotations_handles_only_json(self):
        # parsed_description correctly handles situations where there
        # is no plaintext, only JSON in the description.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = '{"foo": "bar"}'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('', text_before.strip())
        self.assertEqual('', text_after)

    def test_parsing_annotations_handles_multi_line_json(self):
        # parsed_description correctly handles situations where there
        # is multi-line JSON in the description.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{\n  "foo":\n    "bar"\n}\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'bar'}, annotations)
        self.assertEqual('abc', text_before.strip())
        self.assertEqual('xyz', text_after)

    def testJson_with_embedded_braces(self):
        # parsed_description correctly handles situations where there
        # are braces embedded in the JSON.
        card = LeankitCard.create(FauxLane("Test"))
        card.description = 'abc\n{"foo": "b{a}r"}\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual({'foo': 'b{a}r'}, annotations)
        self.assertEqual('abc', text_before.strip())
        self.assertEqual('xyz', text_after)

    def test_bad_JSON_handled(self):
        # If there's bad JSON in a card description, parsed_description
        # is set to an empty Record().
        card = LeankitCard.create(FauxLane("Test"))
        card.id = 1
        card.description = 'abc\n{YAY, Bad JSON}\nxyz\n'
        annotations, text_before, text_after = card.parsed_description
        self.assertEqual(Record(), annotations)
        self.assertEqual('abc', text_before.strip())
        self.assertEqual('xyz', text_after)
