# Copyright 2013 Canonical Ltd.
"""A script to take a blueprint's work items and make cards for them."""

import re
import sys

from ConfigParser import ConfigParser
from lp2kanban.bugs2cards import (
    get_arg_parser,
    get_lp,
    parse_config,
    LaunchpadUsersForBoard,
    )
from lp2kanban.kanban import (
    LeankitKanban,
    Record,
    )


WORK_ITEM_PATTERN = (
    "(\[(?P<assignee>.*)\])?\s*(?P<text>.+)\s*:\s*(?P<status>[A-Z]+)")
BLUEPRINT_URL_PATTERN = (
    "https://blueprints.launchpad.net/(?P<project>.+)/\+spec/(?P<spec>.+)/?")


def process_args():
    parser = get_arg_parser()
    parser.add_argument(
        '-t', '--target-lane', required=True,
        help="The path to the Kanban board lane into which new cards will "
             "be placed")
    parser.add_argument(
        '-s', '--spec-url', required=True,
        help="The URL of the Launchpad blueprint from which cards will be "
             "created.")
    parser.add_argument(
        '--dry-run', required=False, action='store_const', const=True,
        default=False,
        help="If set, no changes will be made to the Kanban board.")
    return parser.parse_args()


def parse_work_items(lp, project, spec):
    project = lp.projects[project]
    blueprint = project.getSpecification(name=spec)
    work_items = set()
    work_item_pattern = re.compile(WORK_ITEM_PATTERN)
    for line in blueprint.workitems_text.split("\n"):
        match = work_item_pattern.match(line)
        if match is None:
            continue
        work_items.add(
            Record(
                blueprint=blueprint.self_link.replace(
                    "api.launchpad.net/devel", "blueprints.launchpad.net"),
                **match.groupdict()))
    return work_items


def work_items_to_cards(work_items, board, lane, lp_users, dry_run=False):
    cards = []
    for work_item in work_items:
        assigned_user = lp_users.lp_to_kanban.get(
            work_item.assignee, None)
        if assigned_user is not None:
            assigned_user_id = assigned_user.id
        else:
            assigned_user_id = 0
        print "Creating card {title} [{assignee}]".format(
            title=work_item.text, assignee=assigned_user_id)
        card = lane.addCard()
        card.title = work_item.text
        card.description_annotations.blueprint_url = work_item.blueprint
        card.type_id = board.default_cardtype.id
        card.description = "Synced from work item"
        card.assigned_user_id = assigned_user_id
        if dry_run:
            card.id = 0
        else:
            card.save()
        cards.append(card)
    return cards


def main(argv):
    args = process_args()
    config = ConfigParser()
    config.read(args.config)
    global_config, boards_config = parse_config(config)

    match = re.match(BLUEPRINT_URL_PATTERN, args.spec_url)
    if match is None:
        print "Spec URL is not a Launchpad blueprint."
        sys.exit(1)

    if not args.dry_run:
        should_continue = raw_input(
            "WARNING: This script is not well-tested and will mess with "
            "your Kanban board; use it at your own risk. Continue? [y/N] ")
        if should_continue.lower() != "y":
            print "Exiting."
            sys.exit(0)

    match_dict = match.groupdict()
    user = args.email or global_config['user']
    password = args.password or global_config['password']
    kanban = LeankitKanban(global_config['account'], user, password)
    board = kanban.getBoard(title=args.board)
    lp = get_lp(boards_config[args.board])
    lp_users = LaunchpadUsersForBoard(lp, board)
    lane = board.getLaneByPath(args.target_lane, ignorecase=True)
    if lane is None:
        print(
            "Unable to find target lane {lane}".format(lane=args.target_lane))
        sys.exit(1)
    work_items = parse_work_items(lp, **match_dict)
    if not args.dry_run:
        should_continue = raw_input(
            "{count} cards will be created, whether or not you want them all. "
            "Continue [y/N]?".format(count=len(work_items)))
    if args.dry_run or should_continue.lower() == "y":
        print "Creating cards from work items in lane {lane}.".format(
            lane=args.target_lane)
        cards = work_items_to_cards(
            work_items, board, lane, lp_users, dry_run=args.dry_run)
        print "{count} cards created".format(count=len(cards))
    else:
        print "Aborting."

if __name__ == '__main__':
    main(sys.argv)
